# -*- coding: utf-8 -*-

"""Utilities for handling external registries."""

import json
import logging
import os
from operator import itemgetter
from typing import List, Mapping, Optional, Union

import requests

from ..constants import EnsureEntry

__all__ = [
    'ensure_registry',
]

logger = logging.getLogger(__name__)


def ensure_registry(
    *,
    url: str,
    embedded_key: str,
    id_key: str,
    cache_path: Optional[str] = None,
    mappify: bool = False,
    force_download: bool = False,
    skip_deprecated: bool = False,
    deprecated_key: Optional[str] = None,
) -> Union[List[EnsureEntry], Mapping[str, EnsureEntry]]:
    """Download the registry (works for MIRIAM and OLS) if it doesn't already exist."""
    if not force_download and cache_path is not None and os.path.exists(cache_path):
        with open(cache_path) as file:
            rv = json.load(file)
    else:
        rv = _download_paginated(url, embedded_key=embedded_key)
        rv = sorted(rv, key=itemgetter(id_key))
        if cache_path is not None:
            with open(cache_path, 'w') as file:
                json.dump(rv, file, indent=2, sort_keys=True)

    if skip_deprecated and deprecated_key:
        rv = [
            entry
            for entry in rv
            if not entry.get(deprecated_key, False)
        ]

    if mappify:
        rv = list_to_map(rv, id_key)

    return rv


def list_to_map(entries, key):
    """Turn a list into a map."""
    return {entry[key]: entry for entry in entries}


def _download_paginated(start_url: str, embedded_key: str) -> List[EnsureEntry]:
    results = []
    url = start_url
    while True:
        logger.debug('getting %s', url)
        g = requests.get(url)
        j = g.json()
        r = j['_embedded'][embedded_key]
        results.extend(r)
        links = j['_links']
        if 'next' not in links:
            break
        url = links['next']['href']
    return results



# -*- coding: utf-8 -*-

"""Download registry information from the OLS."""

from typing import Optional

import click
import pandas as pd

from .utils import ensure_registry
from ..constants import BIOREGISTRY_MODULE

__all__ = [
    'OLS_FULL_PATH',
    'OLS_URL',
    'get_ols',
    'get_ols_df',
]

OLS_URL = 'http://www.ebi.ac.uk/ols/api/ontologies'
OLS_FULL_PATH = BIOREGISTRY_MODULE.join(name='ols.json')
OLS_SLIM_PATH = BIOREGISTRY_MODULE.join(name='ols.tsv')


def get_ols(cache_path: Optional[str] = OLS_FULL_PATH, mappify: bool = False, force_download: bool = False):
    """Get the OLS registry."""
    return ensure_registry(
        url=OLS_URL,
        embedded_key='ontologies',
        cache_path=cache_path,
        id_key='ontologyId',
        mappify=mappify,
        force_download=force_download,
    )


def get_ols_df(**kwargs):
    """Get the OLS registry as a pre-processed dataframe."""
    rows = [
        (
            'ols',
            entry['ontologyId'],
            entry['config']['title'],
            entry['config']['annotations'].get('description'),
            entry['config']['annotations'].get('license', [None])[0],
            entry.get('version'),
        )
        for entry in get_ols(**kwargs)
    ]
    df = pd.DataFrame(rows, columns=[
        'registry', 'prefix', 'name', 'description', 'license', 'version',
    ])
    df.to_csv(OLS_SLIM_PATH, sep='\t', index=False)
    return df


@click.command()
def main():
    """Reload the OLS data."""
    get_ols_df(force_download=True)


if __name__ == '__main__':
    main()



# -*- coding: utf-8 -*-

"""Download registry information from the OBO Foundry."""

import json
import os
from operator import itemgetter
from typing import Optional

import click
import pandas as pd
import yaml

from .utils import list_to_map
from ..constants import BIOREGISTRY_MODULE

__all__ = [
    'OBOFOUNDRY_FULL_PATH',
    'OBOFOUNDRY_URL',
    'get_obofoundry',
    'get_obofoundry_df',
]

OBOFOUNDRY_YAML_PATH = BIOREGISTRY_MODULE.join(name='obofoundry.yml')
OBOFOUNDRY_FULL_PATH = BIOREGISTRY_MODULE.join(name='obofoundry.json')
OBOFOUNDRY_SLIM_PATH = BIOREGISTRY_MODULE.join(name='obofoundry.tsv')
OBOFOUNDRY_URL = 'https://raw.githubusercontent.com/OBOFoundry/OBOFoundry.github.io/master/registry/ontologies.yml'


def get_obofoundry(
    cache_path: Optional[str] = OBOFOUNDRY_FULL_PATH,
    mappify: bool = False,
    force_download: bool = False,
    skip_deprecated: bool = False,
):
    """Get the OBO Foundry registry."""
    if not force_download and cache_path is not None and os.path.exists(cache_path):
        with open(cache_path) as file:
            entries = json.load(file)
    else:
        with BIOREGISTRY_MODULE.ensure(url=OBOFOUNDRY_URL).open() as file:
            entries = yaml.full_load(file)['ontologies']

        for entry in entries:
            for key in ('browsers', 'usages', 'depicted_by', 'products'):
                if key in entry:
                    del entry[key]

        # maintain sorted OBO Foundry
        entries = sorted(entries, key=itemgetter('id'))

        if cache_path is not None:
            with open(cache_path, 'w') as file:
                json.dump(entries, file, indent=2)

    if skip_deprecated:
        entries = [
            entry
            for entry in entries
            if not entry.get('is_obsolete', False)
        ]

    if mappify:
        entries = list_to_map(entries, 'id')

    return entries


def get_obofoundry_df(**kwargs):
    """Get the OBO Foundry registry as a pre-processed dataframe."""
    rows = [
        (
            'obofoundry',
            entry['id'],
            entry['title'],
            entry.get('is_obsolete', False),
            entry.get('license', {}).get('label'),
            entry.get('description'),
        )
        for entry in get_obofoundry(**kwargs)
    ]
    df = pd.DataFrame(rows, columns=[
        'registry', 'prefix', 'name',
        'redundant', 'license', 'description',
    ])
    df.to_csv(OBOFOUNDRY_SLIM_PATH, sep='\t', index=False)
    return df


@click.command()
def main():
    """Reload the OBO Foundry data."""
    get_obofoundry_df(force_download=True)


if __name__ == '__main__':
    main()



# -*- coding: utf-8 -*-

"""Linting functions."""

import click

from bioregistry.utils import secho, updater


@updater
def warn_missing_wikidata(registry):
    """Write warnings for entries completely missing a Wikidata property, database, or paper."""
    _warn_missing_key(registry, 'wikidata')


def _warn_missing_key(registry, key):
    prefixes = [
        prefix
        for prefix, entry in registry.items()
        if key not in entry
    ]
    if prefixes:
        secho(f'Missing {key}:')
        for prefix in prefixes:
            click.echo(f'{prefix:15}: {_get_key(registry, prefix, "name")}')


def _get_key(registry, prefix, key):
    return (
        registry[prefix].get(key)
        or registry[prefix].get("miriam", {}).get(key)
        or registry[prefix].get("obofoundry", {}).get(key)
        or registry[prefix].get("ols", {}).get(key)
        or registry[prefix].get("wikidata", {}).get(key)
        or ''
    )


@updater
def sort_registry(registry):
    """Sort the registry."""
    return registry


@click.command()
def lint():
    """Run the lint commands."""
    warn_missing_wikidata()
    sort_registry()


if __name__ == '__main__':
    lint()



# -*- coding: utf-8 -*-

"""Utility functions for the Bioregistry :mod:`flask` app."""

import itertools as itt
from typing import Any, List, Mapping, Optional

from flask import abort, redirect, render_template, url_for

import bioregistry
from bioregistry.constants import BIOREGISTRY_REMOTE_URL


def _get_resource_providers(prefix: str, identifier: str):
    if identifier is None:
        return
    rv = []
    for metaprefix, url in bioregistry.get_providers_list(prefix, identifier):
        if metaprefix == 'default':
            metaprefix = prefix
            name = bioregistry.get_name(prefix)
            homepage = bioregistry.get_homepage(prefix)
        else:
            name = bioregistry.get_registry_name(metaprefix)
            homepage = bioregistry.get_registry_homepage(metaprefix)
        rv.append(dict(
            metaprefix=metaprefix,
            homepage=homepage,
            name=name,
            url=url,
        ))
    return rv


def _get_resource_mapping_rows(prefix: str):
    mappings = bioregistry.get_mappings(prefix)
    if mappings is None:
        return None
    return [
        dict(
            metaprefix=metaprefix,
            xref=xref,
            homepage=bioregistry.get_registry_homepage(metaprefix),
            name=bioregistry.get_registry_name(metaprefix),
            url=bioregistry.get_registry_url(metaprefix, xref),
        )
        for metaprefix, xref in itt.chain(
            [('bioregistry', prefix)],
            mappings.items(),
        )
    ]


def _normalize_prefix_or_404(prefix: str, endpoint: Optional[str] = None):
    try:
        norm_prefix = bioregistry.normalize_prefix(prefix)
    except ValueError:
        norm_prefix = None
    if norm_prefix is None:
        return render_template('resolve_missing_prefix.html', prefix=prefix), 404
    elif endpoint is not None and norm_prefix != prefix:
        return redirect(url_for(endpoint, prefix=norm_prefix))
    return norm_prefix


def _search(q: str) -> List[str]:
    q_norm = q.lower()
    return [
        prefix
        for prefix in bioregistry.read_bioregistry()
        if q_norm in prefix
    ]


def _autocomplete(q: str) -> Mapping[str, Any]:
    r"""Run the autocomplete algorithm.

    :param q: The query string
    :return: A dictionary with the autocomplete results.

    Before completion is of prefix:

    >>> _autocomplete('cheb')
    {"query": "chebi", "reason": "searched prefix", results=["chebi"], success: True, url: None}

    If only prefix is complete:

    >>> _autocomplete('chebi')
    {"query": "chebi", "reason": "matched prefix", results=["chebi"], success: True, url: "https://bioregistry.io/chebi"}

    Not matching the pattern:

    >>> _autocomplete('chebi:NOPE')
    {"identifier": "NOPE", "pattern": "^CHEBI:\\d+$", "prefix": "chebi", "query": "chebi:NOPE", "reason": "failed validation", "success": False, "url": None}

    Matching the pattern:

    >>> _autocomplete('chebi:1234')
    {"identifier":"1234", "pattern": "^CHEBI:\\d+$", "prefix": "chebi", "query": "chebi:1234", "reason": "passed validation", "success": True, "url": "https://bioregistry.io/chebi:1234"}
    """  # noqa: E501
    if ':' not in q:
        url: Optional[str]
        if q in bioregistry.read_bioregistry():
            reason = 'matched prefix'
            url = f'{BIOREGISTRY_REMOTE_URL.rstrip()}/{q}'
        else:
            reason = 'searched prefix'
            url = None
        return dict(
            query=q,
            results=_search(q),
            success=True,
            reason=reason,
            url=url,
        )
    prefix, identifier = q.split(':', 1)
    norm_prefix = bioregistry.normalize_prefix(prefix)
    if norm_prefix is None:
        return dict(
            query=q,
            prefix=prefix,
            identifier=identifier,
            success=False,
            reason='bad prefix',
        )
    pattern = bioregistry.get_pattern(prefix)
    if pattern is None:
        success = True
        reason = 'no pattern'
        url = bioregistry.resolve_identifier._get_bioregistry_link(prefix, identifier)
    elif bioregistry.validate(prefix, identifier):
        success = True
        reason = 'passed validation'
        url = bioregistry.resolve_identifier._get_bioregistry_link(prefix, identifier)
    else:
        success = False
        reason = 'failed validation'
        url = None
    return dict(
        query=q,
        prefix=prefix,
        pattern=pattern,
        identifier=identifier,
        success=success,
        reason=reason,
        url=url,
    )


def _get_identifier(prefix: str, identifier: str) -> Mapping[str, Any]:
    prefix = _normalize_prefix_or_404(prefix)
    if not bioregistry.validate(prefix, identifier):
        return abort(404, f'invalid identifier: {prefix}:{identifier} for pattern {bioregistry.get_pattern(prefix)}')
    providers = bioregistry.get_providers(prefix, identifier)
    if not providers:
        return abort(404, f'no providers available for {prefix}:{identifier}')

    return dict(
        query=dict(prefix=prefix, identifier=identifier),
        providers=providers,
    )



# -*- coding: utf-8 -*-

"""Download Biolink."""

import requests
import yaml

__all__ = [
    'get_biolink',
]

URL = 'https://raw.githubusercontent.com/biolink/biolink-model/master/biolink-model.yaml'


def get_biolink():
    """Get Biolink."""
    res = requests.get(URL)
    data = yaml.safe_load(res.content)
    return {
        prefix: {'formatter': f'{url}$1'}
        for prefix, url in data['prefixes'].items()
    }


if __name__ == '__main__':
    print(get_biolink())



# -*- coding: utf-8 -*-

"""Align MIRIAM with the Bioregistry."""

from ..external.miriam import get_miriam
from ..utils import norm, secho, updater

MIRIAM_KEYS = {
    'id',
    'prefix',
    'pattern',
    'namespaceEmbeddedInLui',
    'name',
    'deprecated',
    'description',
    'sampleId',
}


@updater
def align_miriam(registry):
    """Update MIRIAM references."""
    miriam_id_to_bioregistry_id = {
        entry['miriam']['id']: key
        for key, entry in registry.items()
        if 'miriam' in entry
    }

    miriam_registry = get_miriam(mappify=True)

    miriam_prefix_to_miriam_id = {
        norm(miriam_entry['prefix']): miriam_entry['mirId'].removeprefix('MIR:')
        for miriam_entry in miriam_registry.values()
    }
    for key, entry in registry.items():
        if 'miriam' in entry:
            continue
        miriam_id = miriam_prefix_to_miriam_id.get(norm(key))
        if miriam_id is not None:
            entry['miriam'] = {'id': miriam_id}
            miriam_id_to_bioregistry_id[miriam_id] = key

    for miriam_entry in miriam_registry.values():
        miriam_entry['id'] = miriam_entry['mirId'].removeprefix('MIR:')

        # Get key by checking the miriam.id key
        bioregistry_id = miriam_id_to_bioregistry_id.get(miriam_entry['id'])
        if bioregistry_id is None:
            if miriam_entry.get('deprecated'):
                secho(f'[{miriam_entry["prefix"]}] skipping deprecated')
                continue
            bioregistry_id = miriam_entry['prefix']
            registry[bioregistry_id] = {}

        registry[bioregistry_id]['miriam'] = {
            k: v
            for k, v in miriam_entry.items()
            if k in MIRIAM_KEYS
        }

    return registry


if __name__ == '__main__':
    align_miriam()



# -*- coding: utf-8 -*-

"""User blueprint for the bioregistry web application."""

from typing import Optional

from flask import Blueprint, abort, redirect, render_template, url_for

import bioregistry
from .utils import _get_resource_mapping_rows, _get_resource_providers, _normalize_prefix_or_404

__all__ = [
    'ui_blueprint',
]

ui_blueprint = Blueprint('ui', __name__)


@ui_blueprint.route('/registry/')
def resources():
    """Serve the Bioregistry page."""
    rows = [
        dict(
            prefix=prefix,
            name=bioregistry.get_name(prefix),
            example=bioregistry.get_example(prefix),
            homepage=bioregistry.get_homepage(prefix),
            pattern=bioregistry.get_pattern(prefix),
            namespace_in_lui=bioregistry.namespace_in_lui(prefix),
            banana=bioregistry.get_banana(prefix),
            description=bioregistry.get_description(prefix),
        )
        for prefix in bioregistry.read_bioregistry()
    ]
    return render_template('resources.html', rows=rows)


@ui_blueprint.route('/metaregistry/')
def metaresources():
    """Serve the Bioregistry metaregistry page."""
    return render_template('metaresources.html', rows=bioregistry.read_metaregistry().values())


@ui_blueprint.route('/registry/<prefix>')
def resource(prefix: str):
    """Serve the a Bioregistry entry page."""
    prefix = _normalize_prefix_or_404(prefix, '.' + resource.__name__)
    if not isinstance(prefix, str):
        return prefix
    example = bioregistry.get_example(prefix)
    return render_template(
        'resource.html',
        prefix=prefix,
        name=bioregistry.get_name(prefix),
        example=example,
        mappings=_get_resource_mapping_rows(prefix),
        synonyms=bioregistry.get_synonyms(prefix),
        homepage=bioregistry.get_homepage(prefix),
        pattern=bioregistry.get_pattern(prefix),
        version=bioregistry.get_version(prefix),
        has_terms=bioregistry.has_terms(prefix),
        obo_download=bioregistry.get_obo_download(prefix),
        owl_download=bioregistry.get_owl_download(prefix),
        namespace_in_lui=bioregistry.namespace_in_lui(prefix),
        deprecated=bioregistry.is_deprecated(prefix),
        contact=bioregistry.get_email(prefix),
        banana=bioregistry.get_banana(prefix),
        description=bioregistry.get_description(prefix),
        providers=None if example is None else _get_resource_providers(prefix, example),
    )


@ui_blueprint.route('/metaregistry/<metaprefix>')
def metaresource(metaprefix: str):
    """Serve the a Bioregistry registry page."""
    entry = bioregistry.get_registry(metaprefix)
    if entry is None:
        abort(404, f'Invalid metaprefix: {metaprefix}')

    example_prefix = entry.get('example')
    example_identifier = bioregistry.get_example(example_prefix) if example_prefix else None
    return render_template(
        'metaresource.html',
        metaprefix=metaprefix,
        name=bioregistry.get_registry_name(metaprefix),
        description=bioregistry.get_registry_description(metaprefix),
        homepage=bioregistry.get_registry_homepage(metaprefix),
        download=entry.get('download'),
        formatter=entry.get('formatter'),
        example_prefix=example_prefix,
        example_prefix_url=bioregistry.get_registry_url(metaprefix, example_prefix) if example_prefix else None,
        example_identifier=example_identifier,
        example_curie_url=(
            bioregistry.get_registry_resolve_url(metaprefix, example_prefix, example_identifier)
            if example_prefix and example_identifier else
            None
        ),
        entry=entry,
    )


@ui_blueprint.route('/reference/<prefix>:<path:identifier>')
def reference(prefix: str, identifier: str):
    """Serve the a Bioregistry reference page."""
    return render_template(
        'reference.html',
        prefix=prefix,
        name=bioregistry.get_name(prefix),
        identifier=identifier,
        providers=_get_resource_providers(prefix, identifier),
    )


@ui_blueprint.route('/<prefix>')
@ui_blueprint.route('/<prefix>:<path:identifier>')
def resolve(prefix: str, identifier: Optional[str] = None):
    """Resolve a CURIE.

    The following things can make a CURIE unable to resolve:

    1. The prefix is not registered with the Bioregistry
    2. The prefix has a validation pattern and the identifier does not match it
    3. There are no providers available for the URL
    """  # noqa:DAR101,DAR201
    norm_prefix = bioregistry.normalize_prefix(prefix)
    if norm_prefix is None:
        return render_template('resolve_missing_prefix.html', prefix=prefix, identifier=identifier), 404
    if identifier is None:
        return redirect(url_for('.' + resource.__name__, prefix=norm_prefix))

    pattern = bioregistry.get_pattern(prefix)
    if pattern and not bioregistry.validate(prefix, identifier):
        return render_template(
            'resolve_invalid_identifier.html', prefix=prefix, identifier=identifier, pattern=pattern,
        ), 404

    url = bioregistry.get_link(prefix, identifier, use_bioregistry_io=False)
    if not url:
        return render_template('resolve_missing_providers.html', prefix=prefix, identifier=identifier), 404
    try:
        # TODO remove any garbage characters?
        return redirect(url)
    except ValueError:  # headers could not be constructed
        return render_template('resolve_disallowed_identifier.html', prefix=prefix, identifier=identifier), 404



# -*- coding: utf-8 -*-

"""Command line interface for the bioregistry."""

import click

from .bioportal import get_bioportal
from .go import get_go
from .miriam import get_miriam_df
from .n2t import get_n2t
from .ncbi import get_ncbi
from .obofoundry import get_obofoundry_df
from .ols import get_ols_df
from .wikidata import get_wikidata_registry
from ..utils import secho

__all__ = [
    'download',
]


@click.command()
def download():
    """Download/update the external entries in the Bioregistry."""
    secho('Downloading MIRIAM')
    get_miriam_df(force_download=True)

    secho('Download N2T')
    get_n2t()

    secho('Downloading NCBI')
    get_ncbi()

    secho('Downloading OBO Foundry')
    get_obofoundry_df(force_download=True)

    secho('Downloading OLS')
    get_ols_df(force_download=True)

    secho('Download Wikidata')
    get_wikidata_registry()

    secho('Download GO')
    get_go(force_download=True)

    secho('Download BioPortal')
    get_bioportal(force_download=True)


if __name__ == '__main__':
    download()



# -*- coding: utf-8 -*-

"""Download Prefix Commons."""

import requests

__all__ = [
    'get_prefix_commons',
]

URL = 'https://raw.githubusercontent.com/prefixcommons/biocontext/master/registry/commons_context.jsonld'


def get_prefix_commons():
    """Get Prefix Commons."""
    return {
        prefix: {'formatter': f'{url}$1'}
        for prefix, url in requests.get(URL).json()['@context'].items()
    }



# -*- coding: utf-8 -*-

"""This script compares what's in each resource."""

import datetime
import itertools as itt
import math
import os
import sys
from collections import Counter
from typing import Collection, Set

import click

from bioregistry import (
    get_description, get_email, get_example, get_format, get_homepage, get_name, get_obo_download, get_owl_download,
    get_pattern, get_version, read_bioregistry,
)
from bioregistry.constants import DOCS_IMG
from bioregistry.external import (
    get_biolink, get_bioportal, get_go, get_miriam, get_n2t, get_ncbi, get_obofoundry, get_ols, get_prefix_commons,
    get_wikidata_registry,
)

bioregistry = read_bioregistry()

LICENSES = {
    'None': None,
    'license': None,
    'unspecified': None,
    # CC-BY (4.0)
    'CC-BY 4.0': 'CC-BY',
    'CC BY 4.0': 'CC-BY',
    'https://creativecommons.org/licenses/by/4.0/': 'CC-BY',
    'http://creativecommons.org/licenses/by/4.0/': 'CC-BY',
    'http://creativecommons.org/licenses/by/4.0': 'CC-BY',
    'https://creativecommons.org/licenses/by/3.0/': 'CC-BY',
    'url: http://creativecommons.org/licenses/by/4.0/': 'CC-BY',
    'SWO is provided under a Creative Commons Attribution 4.0 International'
    ' (CC BY 4.0) license (https://creativecommons.org/licenses/by/4.0/).': 'CC-BY',
    # CC-BY (3.0)
    'CC-BY 3.0 https://creativecommons.org/licenses/by/3.0/': 'CC-BY',
    'http://creativecommons.org/licenses/by/3.0/': 'CC-BY',
    'CC-BY 3.0': 'CC-BY',
    'CC BY 3.0': 'CC-BY',
    'CC-BY version 3.0': 'CC-BY',
    # CC-BY (2.0)
    'CC-BY 2.0': 'CC-BY',
    # CC-BY (unversioned)
    'CC-BY': 'CC-BY',
    'creative-commons-attribution-license': 'CC-BY',
    # CC-BY-SA
    'CC-BY-SA': 'CC-BY-SA',
    'https://creativecommons.org/licenses/by-sa/4.0/': 'CC-BY-SA',
    # CC-BY-NC-SA
    'http://creativecommons.org/licenses/by-nc-sa/2.0/': 'CC-BY-NC-SA',
    # CC 0
    'CC-0': 'CC-0',
    'CC0 1.0 Universal': 'CC-0',
    'CC0': 'CC-0',
    'http://creativecommons.org/publicdomain/zero/1.0/': 'CC-0',
    'https://creativecommons.org/publicdomain/zero/1.0/': 'CC-0',
    # Apache 2.0
    'Apache 2.0 License': 'Other',
    'LICENSE-2.0': 'Other',
    'www.apache.org/licenses/LICENSE-2.0': 'Other',
    # Other
    'GNU GPL 3.0': 'Other',
    'hpo': 'Other',
    'Artistic License 2.0': 'Other',
    'New BSD license': 'Other',
}


def _remap_license(k):
    return k and LICENSES.get(k, k)


SINGLE_FIG = (8, 3.5)
TODAY = datetime.datetime.today().strftime('%Y-%m-%d')
WATERMARK_TEXT = f'https://github.com/bioregistry/bioregistry ({TODAY})'


@click.command()
def compare():  # noqa:C901
    """Compare the registries."""
    try:
        import matplotlib.pyplot as plt
        import seaborn as sns
        from matplotlib_venn import venn2
    except ImportError:
        click.secho(
            'Could not import matplotlib dependencies.'
            ' Install bioregistry again with `pip install bioregistry[charts]`.',
            fg='red',
        )
        return sys.exit(1)

    watermark = True
    sns.set_style('whitegrid')

    ###############################################
    # What kinds of licenses are resources using? #
    ###############################################
    licenses, conflicts, obo_has_license, ols_has_license = _get_license_and_conflicts()

    # How many times does the license appear in OLS / OBO Foundry
    fig, ax = plt.subplots(figsize=SINGLE_FIG)
    venn2(
        subsets=(obo_has_license, ols_has_license),
        set_labels=('OBO', 'OLS'),
        set_colors=('red', 'green'),
        ax=ax,
    )
    if watermark:
        ax.text(
            0.5, -0.1, WATERMARK_TEXT, transform=plt.gca().transAxes,
            fontsize=10, color='gray', alpha=0.5,
            ha='center', va='bottom',
        )

    path = os.path.join(DOCS_IMG, 'license_coverage.svg')
    click.echo(f'output to {path}')
    fig.tight_layout()
    fig.savefig(path, dpi=300)
    plt.close(fig)

    fig, ax = plt.subplots(figsize=SINGLE_FIG)
    sns.countplot(x=licenses, ax=ax)
    ax.set_xlabel('License')
    ax.set_ylabel('Count')
    ax.set_yscale('log')
    if watermark:
        fig.text(
            1.0, 0.5, WATERMARK_TEXT,
            fontsize=8, color='gray', alpha=0.5,
            ha='right', va='center', rotation=90,
        )

    path = os.path.join(DOCS_IMG, 'licenses.svg')
    click.echo(f'output to {path}')
    fig.tight_layout()
    fig.savefig(path, dpi=300)
    plt.close(fig)

    ##############################################
    # How many entries have version information? #
    ##############################################
    def _get_has(f):
        return {key for key in bioregistry if f(key)}

    has_wikidata_database = {
        key
        for key, entry in bioregistry.items()
        if 'database' in entry.get('wikidata', {})
    }
    measurements = [
        ('Name', _get_has(get_name)),
        ('Description', _get_has(get_description)),
        ('Version', _get_has(get_version)),
        ('Homepage', _get_has(get_homepage)),
        ('Contact Email', _get_has(get_email)),
        ('Pattern', _get_has(get_pattern)),
        ('Format URL', _get_has(get_format)),
        ('Example', _get_has(get_example)),
        ('Wikidata Database', has_wikidata_database),
        ('OBO', _get_has(get_obo_download)),
        ('OWL', _get_has(get_owl_download)),
    ]

    ncols = 3
    nrows = int(math.ceil(len(measurements) / ncols))
    fig, axes = plt.subplots(ncols=ncols, nrows=nrows)
    for measurement, ax in itt.zip_longest(measurements, axes.ravel()):
        if measurement is None:
            ax.axis('off')
            continue
        label, prefixes = measurement
        ax.pie(
            (len(prefixes), len(bioregistry) - len(prefixes)),
            labels=('Yes', 'No'),
            autopct='%1.f%%',
            startangle=30,
            explode=[0.1, 0],
        )
        ax.set_title(label)
    if watermark:
        fig.text(
            0.5, 0, WATERMARK_TEXT,
            fontsize=8, color='gray', alpha=0.5,
            ha='center', va='bottom',
        )

    path = os.path.join(DOCS_IMG, 'has_attribute.svg')
    click.echo(f'output to {path}')
    fig.tight_layout()
    fig.savefig(path, dpi=300)
    plt.close(fig)

    # -------------------------------------------------------------------- #

    miriam_prefixes = set(get_miriam(skip_deprecated=True, mappify=True))
    ols_prefixes = set(get_ols(mappify=True))
    obofoundry_prefixes = set(get_obofoundry(skip_deprecated=True, mappify=True))
    wikidata_prefixes = set(get_wikidata_registry())
    n2t_prefixes = set(get_n2t())
    go_prefixes = set(get_go(mappify=True))
    bioportal_prefixes = set(get_bioportal(mappify=True))
    prefixcommons_prefixes = set(get_prefix_commons())
    biolink_prefixes = set(get_biolink())
    ncbi_prefixes = set(get_ncbi())

    keys = [
        ('obofoundry', 'OBO Foundry', 'red', obofoundry_prefixes),
        ('ols', 'OLS', 'green', ols_prefixes),
        ('miriam', 'MIRIAM', 'blue', miriam_prefixes),
        ('wikidata', 'Wikidata', 'purple', wikidata_prefixes),
        ('n2t', 'Name-to-Thing', 'orange', n2t_prefixes),
        ('go', 'GO', 'yellow', go_prefixes),
        ('bioportal', 'BioPortal', 'cyan', bioportal_prefixes),
        ('prefixcommons', 'Prefix Commons', 'magenta', prefixcommons_prefixes),
        ('biolink', 'Biolink Model', 'pink', biolink_prefixes),
        ('ncbi', 'NCBI', 'green', ncbi_prefixes),
    ]

    ############################################################
    # How well does the Bioregistry cover the other resources? #
    ############################################################

    ncols = 3
    nrows = int(math.ceil(len(keys) / ncols))
    figsize = (3.25 * ncols, 2.0 * nrows)
    fig, axes = plt.subplots(ncols=ncols, nrows=nrows, figsize=figsize)
    for key, ax in itt.zip_longest(keys, axes.ravel()):
        if key is None:
            ax.axis('off')
            continue
        key, label, color, prefixes = key
        # Remap bioregistry prefixes to match the external
        #  vocabulary, when possible
        bioregistry_remapped = {
            br_entry.get(key, {}).get('prefix', br_key)
            for br_key, br_entry in bioregistry.items()
        }
        venn2(
            subsets=(bioregistry_remapped, prefixes),
            set_labels=('Bioregistry', label),
            set_colors=('grey', color),
            ax=ax,
        )
    if watermark:
        fig.text(
            0.5, 0, WATERMARK_TEXT,
            fontsize=8, color='gray', alpha=0.5,
            ha='center', va='bottom',
        )

    path = os.path.join(DOCS_IMG, 'bioregistry_coverage.svg')
    click.echo(f'output to {path}')
    fig.tight_layout()
    fig.savefig(path, dpi=300)
    plt.close(fig)

    ######################################################
    # What's the overlap between each pair of resources? #
    ######################################################

    pairs = list(itt.combinations(keys, r=2))
    ncols = 4
    nrows = int(math.ceil(len(pairs) / ncols))
    figsize = (3 * ncols, 2.5 * nrows)
    fig, axes = plt.subplots(ncols=ncols, nrows=nrows, figsize=figsize)
    for pair, ax in itt.zip_longest(pairs, axes.ravel()):
        if pair is None:
            ax.axis('off')
            continue
        (l_key, l_label, l_color, l_prefixes), (r_key, r_label, r_color, r_prefixes) = pair
        # Remap external vocabularies to bioregistry
        #  prefixes, when possible
        l_prefixes = _remap(key=l_key, prefixes=l_prefixes)
        r_prefixes = _remap(key=r_key, prefixes=r_prefixes)
        venn2(
            subsets=(l_prefixes, r_prefixes),
            set_labels=(l_label, r_label),
            set_colors=(l_color, r_color),
            ax=ax,
        )
    if watermark:
        fig.text(
            0.5, 0, WATERMARK_TEXT,  # transform=plt.gca().transAxes,
            fontsize=14, color='gray', alpha=0.5,
            ha='center', va='bottom',
        )

    path = os.path.join(DOCS_IMG, 'external_overlap.svg')
    click.echo(f'output to {path}')
    fig.tight_layout()
    fig.savefig(path, dpi=300)
    plt.close(fig)

    ##############################################
    # Histogram of how many xrefs each entry has #
    ##############################################
    xref_counts = [
        sum(
            key in entry
            for key, *_ in keys
        )
        for entry in bioregistry.values()
    ]
    fig, ax = plt.subplots(figsize=SINGLE_FIG)
    sns.barplot(data=sorted(Counter(xref_counts).items()), ci=None, color='blue', alpha=0.4, ax=ax)
    ax.set_xlabel('Number External References')
    ax.set_ylabel('Count')
    ax.set_yscale('log')
    if watermark:
        fig.text(
            1.0, 0.5, WATERMARK_TEXT,
            fontsize=8, color='gray', alpha=0.5,
            ha='right', va='center', rotation=90,
        )

    path = os.path.join(DOCS_IMG, 'xrefs.svg')
    click.echo(f'output to {path}')
    fig.tight_layout()
    fig.savefig(path, dpi=300)
    plt.close(fig)


def _get_license_and_conflicts():
    licenses = []
    conflicts = set()
    obo_has_license, ols_has_license = set(), set()
    for key, entry in bioregistry.items():
        obo_license = _remap_license(entry.get('obofoundry', {}).get('license'))
        if obo_license:
            obo_has_license.add(key)

        ols_license = _remap_license(entry.get('ols', {}).get('license'))
        if ols_license:
            ols_has_license.add(key)

        if not obo_license and not ols_license:
            licenses.append('None')
        if obo_license and not ols_license:
            licenses.append(obo_license)
        elif not obo_license and ols_license:
            licenses.append(ols_license)
        elif obo_license == ols_license:
            licenses.append(obo_license)
        else:  # different licenses!
            licenses.append(ols_license)
            licenses.append(obo_license)
            conflicts.add(key)
            print(f'[{key}] Conflicting licenses- {obo_license} and {ols_license}')
            continue
    return licenses, conflicts, obo_has_license, ols_has_license


def _remap(*, key: str, prefixes: Collection[str]) -> Set[str]:
    br_external_to = {
        br_entry[key]['prefix']: br_id
        for br_id, br_entry in bioregistry.items()
        if key in br_entry and 'prefix' in br_entry[key]
    }
    return {
        br_external_to.get(prefix, prefix)
        for prefix in prefixes
    }


if __name__ == '__main__':
    compare()



# -*- coding: utf-8 -*-

"""Tests for data integrity."""

import datetime
import logging
import unittest

import bioregistry
from bioregistry.resolve import EMAIL_RE, _get_prefix_key, get_identifiers_org_prefix

logger = logging.getLogger(__name__)


class TestDuplicates(unittest.TestCase):
    """Tests for duplicates."""

    def setUp(self) -> None:
        """Set up the test case."""
        self.registry = bioregistry.read_bioregistry()

    def test_names(self):
        """Test that all entries have a name."""
        for prefix, entry in self.registry.items():
            with self.subTest(prefix=prefix):
                self.assertFalse(
                    'name' not in entry
                    and 'name' not in entry.get('miriam', {})
                    and 'name' not in entry.get('ols', {})
                    and 'name' not in entry.get('obofoundry', {}),
                    msg=f'{prefix} is missing a name',
                )

    def test_name_expansions(self):
        """Test that default names are not capital acronyms."""
        for prefix in bioregistry.read_bioregistry():
            if bioregistry.is_deprecated(prefix):
                continue
            entry = bioregistry.get(prefix)
            if 'name' in entry:
                continue
            name = bioregistry.get_name(prefix)
            if prefix == name.lower() and name.upper() == name:
                with self.subTest(prefix=prefix):
                    self.fail(msg=f'{prefix} acronym ({name}) is not expanded')

            if '.' in prefix and prefix.split('.')[0] == name.lower():
                with self.subTest(prefix=prefix):
                    self.fail(msg=f'{prefix} acronym ({name}) is not expanded')

    def test_homepage_http(self):
        """Test that all homepages start with http."""
        for prefix in bioregistry.read_bioregistry():
            homepage = bioregistry.get_homepage(prefix)
            if homepage is None or homepage.startswith('http'):
                continue
            with self.subTest(prefix=prefix):
                self.fail(msg=f'malformed homepage: {homepage}')

    @unittest.skip('some data needs to be updated from the source. solution proposed at '
                   'https://github.com/OBOFoundry/OBOFoundry.github.io/pull/1480/files')
    def test_email(self):
        """Test that the email getter returns valid email addresses."""
        for prefix in bioregistry.read_bioregistry():
            email = _get_prefix_key(prefix, 'contact', ('obofoundry', 'ols'))
            if email is None or EMAIL_RE.match(email):
                continue
            with self.subTest(prefix=prefix):
                self.fail(msg=f'bad email: {email}')

    def test_no_redundant_acronym(self):
        """Test that there is no redundant acronym in the name.

        For example, "Amazon Standard Identification Number (ASIN)" is a problematic
        name for prefix "asin".
        """
        for prefix in bioregistry.read_bioregistry():
            if bioregistry.is_deprecated(prefix):
                continue
            entry = bioregistry.get(prefix)
            if 'name' in entry:
                continue
            name = bioregistry.get_name(prefix)

            try:
                _, rest = name.rstrip(')').rsplit('(', 1)
            except ValueError:
                continue
            if rest.lower() == prefix.lower():
                with self.subTest(prefix=prefix):
                    self.fail(msg=f'{prefix} has redundany acronym in name "{name}"')

    def test_format_urls(self):
        """Test that entries with a format URL are formatted right (yo dawg)."""
        for prefix, entry in self.registry.items():
            url = entry.get('url')
            if not url:
                continue
            with self.subTest(prefix=prefix):
                self.assertIn('$1', url, msg=f'{prefix} format does not have a $1')

    def test_patterns(self):
        """Test that all prefixes are norm-unique."""
        for prefix, entry in self.registry.items():
            pattern = entry.get('pattern')
            if pattern is None:
                continue
            with self.subTest(prefix=prefix):
                self.assertTrue(pattern.startswith('^'), msg=f'{prefix} pattern {pattern} should start with ^')
                self.assertTrue(pattern.endswith('$'), msg=f'{prefix} pattern {pattern} should end with $')

                # Check that it's the same as external definitions
                for key in ('miriam', 'wikidata'):
                    external_pattern = entry.get('key', {}).get('pattern')
                    if external_pattern:
                        self.assertEqual(pattern, external_pattern, msg=f'{prefix}: {key} pattern not same')

    def test_examples(self):
        """Test that all entries have examples."""
        for prefix, entry in self.registry.items():
            if 'pattern' not in entry:  # TODO remove this later
                continue
            with self.subTest(prefix=prefix):
                msg = f'{prefix} is missing an example local identifier'
                if 'ols' in entry:
                    msg += f'\nSee: https://www.ebi.ac.uk/ols/ontologies/{entry["ols"]["prefix"]}/terms'
                self.assertIsNotNone(bioregistry.get_example(prefix), msg=msg)

    def test_examples_pass_patterns(self):
        """Test that all examples pass the patterns."""
        for prefix, entry in self.registry.items():
            pattern = bioregistry.get_pattern_re(prefix)
            example = bioregistry.get_example(prefix)
            if pattern is None or example is None:
                continue

            if bioregistry.namespace_in_lui(prefix):
                miriam_prefix = get_identifiers_org_prefix(prefix)
                if entry.get('namespace.capitalized') or 'obofoundry' in entry:
                    miriam_prefix = miriam_prefix.upper()
                example = f'{miriam_prefix}:{example}'
            if bioregistry.validate(prefix, example):
                continue
            with self.subTest(prefix=prefix):
                self.assertRegex(example, pattern)

    def test_ols_versions(self):
        """Test that all OLS entries have a version annotation on them."""
        for bioregistry_id, bioregistry_entry in self.registry.items():
            ols = bioregistry_entry.get('ols')
            if not ols:
                continue

            version = ols.get('version')
            if version is None:
                logger.warning('[%s] missing version', bioregistry_id)
                continue

            with self.subTest(prefix=bioregistry_id):
                if version != version.strip():
                    logger.warning('Extra whitespace in %s', bioregistry_id)
                    version = version.strip()

                version_prefix = bioregistry_entry.get('ols_version_prefix')
                if version_prefix:
                    self.assertTrue(version.startswith(version_prefix))
                    version = version[len(version_prefix):]

                if bioregistry_entry.get('ols_version_suffix_split'):
                    version = version.split()[0]

                version_suffix = bioregistry_entry.get('ols_version_suffix')
                if version_suffix:
                    self.assertTrue(version.endswith(version_suffix))
                    version = version[:-len(version_suffix)]

                version_type = bioregistry_entry.get('ols_version_type')
                version_date_fmt = bioregistry_entry.get('ols_version_date_format')
                self.assertTrue(
                    version_type is not None or version_date_fmt is not None,
                    msg=f'missing either a ``ols_version_type`` or ``ols_version_date_format`` for date: {version}',
                )

                if version_date_fmt:
                    if version_date_fmt in {"%Y-%d-%m"}:
                        logger.warning('Confusing date format for %s (%s)', bioregistry_id, version_date_fmt)
                    try:
                        version = datetime.datetime.strptime(version, version_date_fmt)
                    except ValueError:
                        logger.warning('Wrong format for %s (%s)', bioregistry_id, version)



# -*- coding: utf-8 -*-

"""Utilities for normalizing prefixes."""

import datetime
import logging
import re
from functools import lru_cache
from textwrap import dedent
from typing import Any, Mapping, Optional, Sequence, Set, Tuple, Union

from .utils import read_bioregistry, read_metaregistry

__all__ = [
    'get',
    'get_name',
    'get_description',
    'get_mappings',
    'get_synonyms',
    'get_pattern',
    'get_pattern_re',
    'namespace_in_lui',
    'get_format',
    'get_example',
    'has_terms',
    'is_deprecated',
    'get_email',
    'get_homepage',
    'get_obo_download',
    'get_owl_download',
    'parse_curie',
    'normalize_prefix',
    'get_version',
    'get_versions',
    # Metaregistry stuff
    'get_registry',
    'get_registry_name',
    'get_registry_url',
    'get_registry_homepage',
]

logger = logging.getLogger(__name__)

# not a perfect email regex, but close enough
EMAIL_RE = re.compile(r'^(\w|\.|\_|\-)+[@](\w|\_|\-|\.)+[.]\w{2,5}$')


def get(prefix: str) -> Optional[Mapping[str, Any]]:
    """Get the Bioregistry entry for the given prefix.

    :param prefix: The prefix to look up, which is normalized with :func:`normalize_prefix`
        before lookup in the Bioregistry
    :returns: The Bioregistry entry dictionary, which includes several keys cross-referencing
        other registries when available.
    """
    return read_bioregistry().get(normalize_prefix(prefix))


def get_registry(metaprefix: str) -> Optional[Mapping[str, Any]]:
    """Get the metaregistry entry for the given prefix."""
    return read_metaregistry().get(metaprefix)


def get_registry_name(metaprefix: str) -> Optional[str]:
    """Get the metaregistry name for the given prefix, if it's available."""
    registry = get_registry(metaprefix)
    if registry is None:
        return None
    return registry['name']


def get_registry_homepage(metaprefix: str) -> Optional[str]:
    """Get the URL for the registry, if available."""
    return _get_registry_key(metaprefix, 'homepage')


def get_registry_description(metaprefix: str) -> Optional[str]:
    """Get the description for the registry, if available."""
    return _get_registry_key(metaprefix, 'description')


def get_registry_example(metaprefix: str) -> Optional[str]:
    """Get an example for the registry, if available."""
    return _get_registry_key(metaprefix, 'example')


def _get_registry_key(metaprefix: str, key: str):
    registry = get_registry(metaprefix)
    if registry is None:
        return None
    return registry.get(key)


def get_registry_url(metaprefix: str, prefix: str) -> Optional[str]:
    """Get the URL for the resource inside registry, if available."""
    entry = get_registry(metaprefix)
    if entry is None:
        return None
    formatter = entry.get('formatter')
    if formatter is None:
        return None
    return formatter.replace('$1', prefix)


def get_name(prefix: str) -> Optional[str]:
    """Get the name for the given prefix, it it's available."""
    return _get_prefix_key(prefix, 'name', ('obofoundry', 'ols', 'wikidata', 'go', 'ncbi', 'bioportal', 'miriam'))


def get_description(prefix: str) -> Optional[str]:
    """Get the description for the given prefix, if available."""
    return _get_prefix_key(prefix, 'description', ('miriam', 'ols', 'obofoundry', 'wikidata'))


def get_mappings(prefix: str) -> Optional[Mapping[str, str]]:
    """Get the mappings to external registries, if available."""
    entry = get(prefix)
    if entry is None:
        return None
    rv = {}
    for key in read_metaregistry():
        if key not in entry:
            continue
        if key != 'wikidata':
            rv[key] = entry[key]['prefix']
        else:
            value = entry[key].get('property')
            if value is not None:
                rv['wikidata'] = value
    return rv


def get_synonyms(prefix: str) -> Optional[Set[str]]:
    """Get the synonyms for a given prefix, if available."""
    entry = get(prefix)
    if entry is None:
        return None
    # TODO aggregate even more from xrefs
    return entry.get('synonyms')


def _get_prefix_key(prefix: str, key: str, sources: Sequence[str]):
    entry = get(prefix)
    if entry is None:
        return None
    rv = entry.get(key)
    if rv is not None:
        return rv
    for source in sources:
        rv = entry.get(source, {}).get(key)
        if rv is not None:
            return rv
    return None


def get_pattern(prefix: str) -> Optional[str]:
    """Get the pattern for the given prefix, if it's available.

    :param prefix: The prefix to look up, which is normalized with :func:`normalize_prefix`
        before lookup in the Bioregistry
    :returns: The pattern for the prefix, if it is available, using the following order of preference:
        1. Custom
        2. MIRIAM
        3. Wikidata
    """
    return _get_prefix_key(prefix, 'pattern', ('miriam', 'wikidata'))


@lru_cache()
def get_pattern_re(prefix: str):
    """Get the compiled pattern for the given prefix, if it's available."""
    pattern = get_pattern(prefix)
    if pattern is None:
        return None
    return re.compile(pattern)


def namespace_in_lui(prefix: str) -> Optional[bool]:
    """Check if the namespace should appear in the LUI."""
    return _get_prefix_key(prefix, 'namespaceEmbeddedInLui', ('miriam',))


def get_identifiers_org_prefix(prefix: str) -> Optional[str]:
    """Get the identifiers.org prefix if available."""
    return _get_mapped_prefix(prefix, 'miriam')


def get_n2t_prefix(prefix: str) -> Optional[str]:
    """Get the name-to-thing prefix if available."""
    return _get_mapped_prefix(prefix, 'n2t')


def get_bioportal_prefix(prefix: str) -> Optional[str]:
    """Get the Bioportal prefix if available."""
    return _get_mapped_prefix(prefix, 'bioportal')


def get_obofoundry_prefix(prefix: str) -> Optional[str]:
    """Get the OBO Foundry prefix if available."""
    return _get_mapped_prefix(prefix, 'obofoundry')


def get_ols_prefix(prefix: str) -> Optional[str]:
    """Get the OLS prefix if available."""
    return _get_mapped_prefix(prefix, 'ols')


def _get_mapped_prefix(prefix: str, external: str) -> Optional[str]:
    entry = get(prefix)
    if entry is None:
        return None
    return entry.get(external, {}).get('prefix')


def get_banana(prefix: str) -> Optional[str]:
    """Get the optional redundant prefix to go before an identifier."""
    entry = get(prefix)
    if entry is None:
        return None
    return entry.get('banana')


def get_format(prefix: str) -> Optional[str]:
    """Get the URL format string for the given prefix, if it's available."""
    entry = get(prefix)
    if entry is None:
        return None
    url = entry.get('url')
    if url is not None:
        return url
    miriam_id = get_identifiers_org_prefix(prefix)
    if miriam_id is not None:
        if namespace_in_lui(prefix):
            # not exact solution, some less common ones don't use capitalization
            # align with the banana solution
            miriam_id = miriam_id.upper()
        return f'https://identifiers.org/{miriam_id}:$1'
    ols_id = entry.get('ols', {}).get('prefix')
    if ols_id is not None:
        purl = f'http://purl.obolibrary.org/obo/{ols_id.upper()}_$1'
        return f'https://www.ebi.ac.uk/ols/ontologies/{ols_id}/terms?iri={purl}'
    return None


def get_example(prefix: str) -> Optional[str]:
    """Get an example identifier, if it's available."""
    entry = get(prefix)
    if entry is None:
        return None
    example = entry.get('example')
    if example is not None:
        return example
    miriam_example = entry.get('miriam', {}).get('sampleId')
    if miriam_example is not None:
        return miriam_example
    example = entry.get('ncbi', {}).get('example')
    if example is not None:
        return example
    return None


def has_terms(prefix: str) -> bool:
    """Check if the prefix is specifically noted to not have terms."""
    entry = get(prefix)
    if entry is None:
        return True
    return not entry.get('no_own_terms', False)


def is_deprecated(prefix: str) -> bool:
    """Return if the given prefix corresponds to a deprecated resource."""
    entry = get(prefix)
    if entry is None:
        return False
    if 'deprecated' in entry:
        return entry['deprecated']
    for key in ('obofoundry', 'ols', 'miriam'):
        if key in entry and 'deprecated' in entry[key]:
            return entry[key]['deprecated']
    return False


def get_email(prefix: str) -> Optional[str]:
    """Return the contact email, if available."""
    rv = _get_prefix_key(prefix, 'contact', ('obofoundry', 'ols'))
    if rv and not EMAIL_RE.match(rv):
        logger.warning('[%s] invalid email address listed: %s', prefix, rv)
        return None
    return rv


def get_homepage(prefix: str) -> Optional[str]:
    """Return the homepage, if available."""
    return _get_prefix_key(prefix, 'homepage', ('obofoundry', 'ols', 'n2t', 'wikidata', 'go', 'ncbi'))


def get_obo_download(prefix: str) -> Optional[str]:
    """Get the download link for the latest OBO file."""
    entry = get(prefix)
    if entry is None:
        return None
    return entry.get('obofoundry', {}).get('download.obo')


def get_owl_download(prefix: str) -> Optional[str]:
    """Get the download link for the latest OWL file."""
    entry = get(prefix)
    if entry is None:
        return None
    return entry.get('ols', {}).get('version.iri')


def is_provider(prefix: str) -> bool:
    """Get if the prefix is a provider.

    :param prefix: The prefix to look up
    :returns: if the prefix is a provider

    >>> assert not is_provider('pdb')
    >>> assert is_provider('validatordb')
    """
    entry = get(prefix)
    if entry is None:
        return False
    return entry.get('type') == 'provider'


def get_provides_for(prefix: str) -> Optional[str]:
    """Get the resource that the given prefix provides for, or return none if not a provider.

    :param prefix: The prefix to look up
    :returns: The prefix of the resource that the given prefix provides for, if it's a provider

    >>> assert get_provides_for('pdb') is None
    >>> assert 'pdb' == get_provides_for('validatordb')
    """
    entry = get(prefix)
    if entry is None:
        return None
    return entry.get('provides')


def parse_curie(curie: str) -> Union[Tuple[str, str], Tuple[None, None]]:
    """Parse a CURIE, normalizing the prefix and identifier if necessary.

    :param curie: A compact URI (CURIE) in the form of <prefix:identifier>
    :returns: A tuple of the prefix, identifier. If not parsable, returns a tuple of None, None

    Parse canonical CURIE
    >>> parse_curie('go:1234')
    ('go', '1234')

    Normalize prefix
    >>> parse_curie('GO:1234')
    ('go', '1234')

    Address banana problem
    >>> parse_curie('GO:GO:1234')
    ('go', '1234')
    """
    try:
        prefix, identifier = curie.split(':', 1)
    except ValueError:
        return None, None

    # remove redundant prefix
    if identifier.casefold().startswith(f'{prefix.casefold()}:'):
        identifier = identifier[len(prefix) + 1:]

    norm_prefix = normalize_prefix(prefix)
    if not norm_prefix:
        return None, None
    return norm_prefix, identifier


def normalize_prefix(prefix: str) -> Optional[str]:
    """Get the normalized prefix, or return None if not registered.

    :param prefix: The prefix to normalize, which could come from Bioregistry,
        OBO Foundry, OLS, or any of the curated synonyms in the Bioregistry
    :returns: The canonical Bioregistry prefix, it could be looked up. This
        will usually take precedence: MIRIAM, OBO Foundry / OLS, Custom except
        in a few cases, such as NCBITaxon.

    This works for synonym prefixes, like:

    >>> assert 'ncbitaxon' == normalize_prefix('taxonomy')

    This works for common mistaken prefixes, like:

    >>> assert 'pubchem.compound' == normalize_prefix('pubchem')

    This works for prefixes that are often written many ways, like:

    >>> assert 'eccode' == normalize_prefix('ec-code')
    >>> assert 'eccode' == normalize_prefix('EC_CODE')
    """
    return _synonym_to_canonical().get(prefix)


def _norm(s: str) -> str:
    """Normalize a string for dictionary key usage."""
    rv = s.casefold().lower()
    for x in ' .-_./':
        rv = rv.replace(x, '')
    return rv


class NormDict(dict):
    def __setitem__(self, key: str, value: str) -> None:
        norm_key = _norm(key)
        if value is None:
            raise ValueError(f'Tried to add empty value for {key}/{norm_key}')
        if norm_key in self and self[norm_key] != value:
            raise KeyError(f'Tried to add {norm_key}/{value} when already had {norm_key}/{self[norm_key]}')
        super().__setitem__(norm_key, value)

    def __getitem__(self, item: str) -> str:
        return super().__getitem__(_norm(item))

    def get(self, key: str, default=None) -> str:
        return super().get(_norm(key), default)


@lru_cache(maxsize=1)
def _synonym_to_canonical() -> NormDict:
    """Return a mapping from several variants of each synonym to the canonical namespace."""
    norm_synonym_to_key = NormDict()

    for bioregistry_id, entry in read_bioregistry().items():
        norm_synonym_to_key[bioregistry_id] = bioregistry_id

        for external in ('miriam', 'ols', 'wikidata', 'obofoundry', 'go'):
            if external in entry and 'prefix' in entry[external]:
                norm_synonym_to_key[entry[external]['prefix']] = bioregistry_id

        for synonym in entry.get('synonyms', []):
            norm_synonym_to_key[synonym] = bioregistry_id

    return norm_synonym_to_key


def get_version(prefix: str) -> Optional[str]:
    """Get the version."""
    norm_prefix = normalize_prefix(prefix)
    if norm_prefix is None:
        return None
    return get_versions().get(norm_prefix)


@lru_cache(maxsize=1)
def get_versions() -> Mapping[str, str]:
    """Get a map of prefixes to versions."""
    rv = {}

    for bioregistry_id, bioregistry_entry in read_bioregistry().items():
        if 'ols' not in bioregistry_entry:
            continue
        version = bioregistry_entry['ols'].get('version')
        if version is None:
            logger.warning('[%s] missing version. Contact: %s', bioregistry_id, get_email(bioregistry_id))
            continue

        version = _clean_version(bioregistry_id, version, bioregistry_entry=bioregistry_entry)
        version_type = bioregistry_entry.get('ols_version_type')
        version_date_fmt = bioregistry_entry.get('ols_version_date_format')

        if version_date_fmt:
            if version_date_fmt in {"%Y-%d-%m"}:
                logger.warning(
                    '[%s] confusing date format: %s. Contact: %s', bioregistry_id, version_date_fmt,
                    get_email(bioregistry_id),
                )
            try:
                version = datetime.datetime.strptime(version, version_date_fmt).strftime('%Y-%m-%d')
            except ValueError:
                logger.warning('[%s] wrong format for version %s', bioregistry_id, version)
        elif not version_type:
            logger.warning('[%s] no type for version %s', bioregistry_id, version)

        rv[bioregistry_id] = version

    return rv


def _clean_version(
    bioregistry_id: str,
    version: str,
    *,
    bioregistry_entry: Optional[Mapping[str, Any]] = None,
) -> str:
    if bioregistry_entry is None:
        bioregistry_entry = get(bioregistry_id)
    if bioregistry_entry is None:
        raise ValueError

    if version != version.strip():
        logger.warning(
            '[%s] extra whitespace in version: %s. Contact: %s',
            bioregistry_id, version, get_email(bioregistry_id),
        )
        version = version.strip()

    version_prefix = bioregistry_entry.get('ols_version_prefix')
    if version_prefix:
        if not version.startswith(version_prefix):
            raise ValueError(dedent(f'''\
            [{bioregistry_id}] version "{version}" does not start with prefix "{version_prefix}".
            Update the ["{bioregistry_id}"]["ols_version_prefix"] entry.
            '''))
        version = version[len(version_prefix):]

    if bioregistry_entry.get('ols_version_suffix_split'):
        version = version.split()[0]

    version_suffix = bioregistry_entry.get('ols_version_suffix')
    if version_suffix:
        if not version.endswith(version_suffix):
            raise ValueError(f'[{bioregistry_id}] version {version} does not end with prefix {version_suffix}')
        version = version[:-len(version_suffix)]

    return version



# -*- coding: utf-8 -*-

"""Query, download, and format Wikidata as a registry."""

import logging
from typing import Iterable, Tuple

import pandas as pd

from bioregistry.constants import BIOREGISTRY_MODULE
from bioregistry.external.miriam import get_miriam
from bioregistry.utils import query_wikidata

logger = logging.getLogger(__name__)

DATABASES_PATH = BIOREGISTRY_MODULE.join(name='wikidata_databases.tsv')
MIRIAM_PATH = BIOREGISTRY_MODULE.join(name='wikidata_miriam.tsv')


def get_database() -> pd.DataFrame:
    """Get the databases dataframe."""
    df = pd.DataFrame(iter_database(), columns=['database_id', 'database_label', 'prop_id', 'prop_label'])
    df.to_csv(DATABASES_PATH, sep='\t', index=False)
    return df


def iter_database() -> Iterable[Tuple[str, str, str, str]]:
    """Iterate over database-property pairs from Wikidata."""
    query = """
    SELECT ?database ?databaseLabel ?prop ?propLabel
    WHERE
    {
      ?database wdt:P31 wd:Q4117139 .
      ?database wdt:P1687 ?prop .
      SERVICE wikibase:label { bd:serviceParam wikibase:language "[AUTO_LANGUAGE],en". }
    }
    """
    for d in query_wikidata(query):
        database_id = d['database']['value'].split('/')[-1]
        database_label = d['databaseLabel']['value']
        prop_id = d['prop']['value'].split('/')[-1]
        prop_label = d['propLabel']['value']
        yield database_id, database_label, prop_id, prop_label


def get_miriam_mappings() -> pd.DataFrame:
    """Get MIRIAM-Wikidata mappings."""
    df = pd.DataFrame(iter_wikidata_mappings(), columns=['prop_id', 'prop_label', 'miriam_id', 'miriam_label'])
    df.to_csv(MIRIAM_PATH, sep='\t', index=False)
    return df


def iter_wikidata_mappings() -> Iterable[Tuple[str, str, str, str]]:
    """Iterate over Wikidata xrefs."""
    miriam = get_miriam(mappify=True)

    query = """SELECT ?item ?itemLabel ?miriam
    WHERE
    {
      ?item wdt:P4793 ?miriam .
      SERVICE wikibase:label { bd:serviceParam wikibase:language "[AUTO_LANGUAGE],en". }
    }
    """
    # P4793 identifiers.org prefix
    #

    wikidata_to_miriam = {
        'wormbase': 'wb',
        'flybase': 'fb',
    }

    for d in query_wikidata(query):
        wikidata_id = d['item']['value'].split('/')[-1]
        wikidata_label = d['itemLabel']['value']
        miriam_label = d['miriam']['value'].lower()
        miriam_label = wikidata_to_miriam.get(miriam_label, miriam_label)

        miriam_entry = miriam.get(miriam_label)
        if miriam_entry is None:
            logger.debug('MISSING %s %s %s', wikidata_id, wikidata_label, miriam_label)
            continue

        miriam_id = miriam_entry['mirId'].removeprefix('MIR:')
        yield wikidata_id, wikidata_label, miriam_id, miriam_label


def get_wikidata_registry():
    """Get the wikidata registry."""
    m = get_miriam_mappings()
    get_database()
    return set(m['miriam_label'])


if __name__ == '__main__':
    get_wikidata_registry()



# -*- coding: utf-8 -*-

"""Download the NCBO BioPortal registry.

Get an API key by logging up, signing in, and navigating to https://bioportal.bioontology.org/account.
"""

import json

import click
import pystow
import requests
from more_click import verbose_option

from bioregistry.constants import BIOREGISTRY_MODULE
from bioregistry.external.utils import list_to_map

__all__ = [
    'get_bioportal',
]

BIOPORTAL_PATH = BIOREGISTRY_MODULE.join(name='bioportal.json')
BIOPORTAL_API_KEY = pystow.get_config('bioportal', 'api_key')
BASE_URL = 'https://data.bioontology.org'


def query(url: str, **params) -> requests.Response:
    """Query the given endpoint on BioPortal."""
    if BIOPORTAL_API_KEY is None:
        raise ValueError('missing API key for bioportal')
    params.setdefault('apikey', BIOPORTAL_API_KEY)
    return requests.get(f'{BASE_URL}/{url}', params=params)


def get_bioportal(force_download: bool = True, mappify: bool = False):
    """Get the BioPortal registry."""
    if BIOPORTAL_PATH.exists() and not force_download:
        with BIOPORTAL_PATH.open() as file:
            entries = json.load(file)
    else:
        # see https://data.bioontology.org/documentation#Ontology
        res = query('ontologies', summaryOnly=False, notes=True)
        entries = res.json()
        for entry in entries:
            for key in ('links', '@context'):
                if key in entry:
                    del entry[key]

        with BIOPORTAL_PATH.open('w') as file:
            json.dump(entries, file, indent=2)

    if mappify:
        try:
            entries = list_to_map(entries, 'acronym')
        except TypeError:
            print(entries)
            raise

    return entries


@click.command()
@verbose_option
def _main():
    get_bioportal()


if __name__ == '__main__':
    _main()



# -*- coding: utf-8 -*-

"""Workflows for downloading and aligning various external registries."""

from .biolink import get_biolink  # noqa:F401
from .bioportal import get_bioportal  # noqa:F401
from .go import get_go  # noqa:F401
from .miriam import get_miriam, get_miriam_df  # noqa:F401
from .n2t import get_n2t  # noqa:F401
from .ncbi import get_ncbi  # noqa:F401
from .obofoundry import get_obofoundry, get_obofoundry_df  # noqa:F401
from .ols import get_ols, get_ols_df  # noqa:F401
from .prefix_commons import get_prefix_commons  # noqa:F401
from .wikidata import get_wikidata_registry  # noqa:F401



# -*- coding: utf-8 -*-

"""Constants and utilities for registries."""

import os
import pathlib
from typing import Any

import pystow

__all__ = [
    'HERE',
    'DATA_DIRECTORY',
    'BIOREGISTRY_PATH',
    'METAREGISTRY_PATH',
    'BIOREGISTRY_MODULE',
    'EnsureEntry',
]

HERE = pathlib.Path(os.path.abspath(os.path.dirname(__file__)))
DATA_DIRECTORY = HERE / 'data'
BIOREGISTRY_PATH = DATA_DIRECTORY / 'bioregistry.json'
METAREGISTRY_PATH = DATA_DIRECTORY / 'metaregistry.json'

BIOREGISTRY_MODULE = pystow.module('bioregistry')
EnsureEntry = Any

DOCS = os.path.abspath(os.path.join(HERE, os.pardir, os.pardir, 'docs'))
DOCS_DATA = os.path.join(DOCS, '_data')
DOCS_IMG = os.path.join(DOCS, 'img')

#: The URL of the remote Bioregistry site
BIOREGISTRY_REMOTE_URL = pystow.get_config('bioregistry', 'url') or 'https://bioregistry.io'

#: Resolution is broken on identifiers.org for the following
IDOT_BROKEN = {
    'gramene.growthstage',
    'oma.hog',
    'obi',
}



# -*- coding: utf-8 -*-

"""Utilities."""

import json
import logging
from datetime import datetime
from functools import lru_cache, wraps
from typing import Any, List, Mapping

import click
import requests

from .constants import BIOREGISTRY_PATH, METAREGISTRY_PATH

logger = logging.getLogger(__name__)


@lru_cache(maxsize=1)
def read_metaregistry() -> Mapping[str, Mapping[str, Any]]:
    """Read the metaregistry as JSON."""
    with open(METAREGISTRY_PATH, encoding='utf-8') as file:
        return {
            entry['prefix']: entry
            for entry in json.load(file)
        }


@lru_cache(maxsize=1)
def read_bioregistry():
    """Read the Bioregistry as JSON."""
    with open(BIOREGISTRY_PATH, encoding='utf-8') as file:
        return json.load(file)


def write_bioregistry(registry):
    """Write to the Bioregistry."""
    with open(BIOREGISTRY_PATH, mode='w', encoding='utf-8') as file:
        json.dump(registry, file, indent=2, sort_keys=True, ensure_ascii=False)


def updater(f):
    """Make a decorator for functions that auto-update the bioregistry."""

    @wraps(f)
    def wrapped():
        registry = read_bioregistry()
        rv = f(registry)
        if rv is not None:
            write_bioregistry(registry)
        return rv

    return wrapped


def norm(s: str) -> str:
    """Normalize a string for dictionary key usage."""
    rv = s.lower()
    for x in ' .-':
        rv = rv.replace(x, '')
    return rv


def secho(s, fg='cyan', bold=True, **kwargs):
    """Wrap :func:`click.secho`."""
    click.echo(f'[{datetime.now().strftime("%H:%M:%S")}] ' + click.style(s, fg=fg, bold=bold, **kwargs))


#: Wikidata SPARQL endpoint. See https://www.wikidata.org/wiki/Wikidata:SPARQL_query_service#Interfacing
WIKIDATA_ENDPOINT = 'https://query.wikidata.org/bigdata/namespace/wdq/sparql'


def query_wikidata(sparql: str) -> List[Mapping[str, Any]]:
    """Query Wikidata's sparql service.

    :param sparql: A SPARQL query string
    :return: A list of bindings
    """
    logger.debug('running query: %s', sparql)
    res = requests.get(WIKIDATA_ENDPOINT, params={'query': sparql, 'format': 'json'})
    res.raise_for_status()
    res_json = res.json()
    return res_json['results']['bindings']



# -*- coding: utf-8 -*-

"""Download registry information from Identifiers.org/MIRIAMs."""

from typing import Optional

import click
import pandas as pd

from .utils import ensure_registry
from ..constants import BIOREGISTRY_MODULE

__all__ = [
    'MIRIAM_FULL_PATH',
    'MIRIAM_URL',
    'get_miriam',
    'get_miriam_df',
]

MIRIAM_FULL_PATH = BIOREGISTRY_MODULE.join(name='miriam.json')
MIRIAM_SLIM_PATH = BIOREGISTRY_MODULE.join(name='miriam.tsv')
MIRIAM_URL = 'https://registry.api.identifiers.org/restApi/namespaces'


def get_miriam(
    cache_path: Optional[str] = MIRIAM_FULL_PATH,
    mappify: bool = False,
    force_download: bool = False,
    skip_deprecated: bool = False,
):
    """Get the MIRIAM registry."""
    return ensure_registry(
        url=MIRIAM_URL,
        embedded_key='namespaces',
        cache_path=cache_path,
        id_key='prefix',
        mappify=mappify,
        force_download=force_download,
        deprecated_key='deprecated',
        skip_deprecated=skip_deprecated,
    )


def get_miriam_df(**kwargs):
    """Get the MIRIAM registry as a pre-processed dataframe."""
    rows = [
        (
            'miriam',
            entry['mirId'],
            entry['prefix'],
            entry['pattern'],
            entry['namespaceEmbeddedInLui'],
            entry['name'],
            entry['deprecated'],
            entry['description'],
            entry['sampleId'],
        )
        for entry in get_miriam(**kwargs)
    ]
    df = pd.DataFrame(rows, columns=[
        'registry', 'identifier', 'prefix',
        'pattern', 'namespaceEmbeddedInLui', 'name', 'deprecated', 'description', 'sampleId',
    ])
    df.to_csv(MIRIAM_SLIM_PATH, sep='\t', index=False)
    return df


@click.command()
def main():
    """Reload the MIRIAM data."""
    get_miriam_df(force_download=True)


if __name__ == '__main__':
    main()



# -*- coding: utf-8 -*-

"""CLI for alignment."""

import click

from .biolink import BiolinkAligner
from .bioportal import BioPortalAligner
from .go import GoAligner
from .miriam import align_miriam
from .n2t import align_n2t
from .ncbi import NcbiAligner
from .obofoundry import align_obofoundry
from .ols import align_ols
from .prefix_commons import PrefixCommonsAligner
from .wikidata import align_wikidata
from ..utils import secho

__all__ = [
    'align',
]


@click.command()
def align():
    """Align all external registries."""
    secho('Aligning MIRIAM')
    align_miriam()

    secho('Aligning N2T')
    align_n2t()

    secho('Aligning NCBI')
    NcbiAligner.align()

    secho('Aligning OBO Foundry')
    align_obofoundry()

    secho('Aligning OLS')
    align_ols()

    secho('Aligning Wikidata')
    align_wikidata()

    secho('Aligning GO')
    GoAligner.align()

    secho('Aligning BioPortal')
    BioPortalAligner.align()

    secho('Aligning Prefix Commons')
    PrefixCommonsAligner.align()

    secho('Aligning Biolink')
    BiolinkAligner.align()


if __name__ == '__main__':
    align()



# -*- coding: utf-8 -*-

"""Align N2T with the Bioregistry."""

from ..external import get_n2t
from ..utils import norm, updater


@updater
def align_n2t(registry):
    """Update N2T references."""
    n2t_id_to_bioregistry_id = {
        entry['n2t']['prefix']: key
        for key, entry in registry.items()
        if 'n2t' in entry
    }
    n2t = get_n2t()
    n2t_norm_prefix_to_prefix = {
        norm(n2t_key): n2t_key
        for n2t_key in n2t
    }
    for bioregistry_id, entry in registry.items():
        if 'n2t' in entry:
            continue
        n2t_id = n2t_norm_prefix_to_prefix.get(norm(bioregistry_id))
        if n2t_id is not None:
            entry['n2t'] = {'prefix': n2t_id}
            n2t_id_to_bioregistry_id[n2t_id] = bioregistry_id

    for n2t_prefix, _n2t_entry in n2t.items():
        bioregistry_id = n2t_id_to_bioregistry_id.get(n2t_prefix)
        if bioregistry_id is None:
            continue
        registry[bioregistry_id]['n2t'] = {
            'prefix': n2t_prefix,
            # TODO add providers? they're not so informative though
            # 'providers': n2t_entry,
        }
        if len(_n2t_entry) == 1:
            registry[bioregistry_id]['n2t']['homepage'] = _n2t_entry[0]['homepage']
            registry[bioregistry_id]['n2t']['name'] = _n2t_entry[0]['title']

    return registry


if __name__ == '__main__':
    align_n2t()



# -*- coding: utf-8 -*-

"""Align the GO with the Bioregistry."""

import json
from typing import Any, Dict, Mapping

from bioregistry.align.utils import Aligner
from bioregistry.constants import DATA_DIRECTORY
from bioregistry.external import get_go

__all__ = [
    'GoAligner',
]

PROCESSING_GO_PATH = DATA_DIRECTORY / 'processing_go.json'


class GoAligner(Aligner):
    """An aligner for the Gene Ontology (GO) registry."""

    key = 'go'
    getter = get_go
    getter_kwargs = {'mappify': True}
    curation_header = 'name', 'description'

    def get_skip(self) -> Mapping[str, str]:
        """Get the skipped GO identifiers."""
        with PROCESSING_GO_PATH.open() as file:
            j = json.load(file)
        return j['skip']

    def prepare_external(self, external_id, external_entry) -> Dict[str, Any]:
        """Prepare GO data to be added to the bioregistry for each GO registry entry."""
        rv = {
            'name': external_entry['name'],
        }

        description = external_entry.get('desription')
        if description:
            rv['description'] = description

        homepages = [
            homepage
            for homepage in external_entry.get('generic_urls', [])
            if not any(
                homepage.startswith(homepage_prefix)
                for homepage_prefix in [
                    'http://purl.obolibrary.org',
                ]
            )
        ]
        if len(homepages) > 1:
            print(external_id, 'multiple homepages', homepages)
        if homepages:
            rv['homepage'] = homepages[0]

        entity_types = external_entry.get('entity_types', [])
        if len(entity_types) > 1:
            print(external_id, 'multiple entity types')
            # TODO handle
        elif len(entity_types) == 1:
            entity_type = entity_types[0]
            formatter = entity_type.get('url_syntax')
            if formatter and not any(
                formatter.startswith(formatter_prefix)
                for formatter_prefix in [
                    'http://purl.obolibrary.org',
                    'https://purl.obolibrary.org',
                ]
            ):
                formatter = formatter.replace('[example_id]', '$1')
                rv['formatter'] = formatter

        if 'synonyms' in external_entry:
            rv['synonyms'] = external_entry['synonyms']

        return rv

    def get_curation_row(self, external_id, external_entry):
        """Prepare curation rows for unaligned GO registry entries."""
        return [external_entry['name'], external_entry.get('description')]


if __name__ == '__main__':
    GoAligner.align()



# -*- coding: utf-8 -*-

"""Command line interface for the bioregistry."""

import os

import click
import yaml
from more_click import make_web_command

from .align.cli import align
from .compare import compare
from .constants import DOCS_DATA
from .external.cli import download
from .generate_warnings_file import warnings
from .lint import lint
from .make_curation_list import curation


@click.group()
@click.version_option()
def main():
    """Run the Bioregistry CLI."""


main.add_command(lint)
main.add_command(compare)
main.add_command(warnings)
main.add_command(make_web_command('bioregistry.app.wsgi:app'))


@main.command()
def copy():
    """Copy the source Bioregistry to the docs folder."""
    from . import read_bioregistry
    registry = read_bioregistry()
    ov = [
        {
            'prefix': prefix,
            **data,
        }
        for prefix, data in registry.items()
    ]
    with open(os.path.join(DOCS_DATA, 'bioregistry.yml'), 'w') as file:
        yaml.dump(ov, file)


@main.command()
def versions():
    """Print the versions."""
    from . import read_bioregistry
    from .resolve import get_versions
    from tabulate import tabulate

    registry = read_bioregistry()
    click.echo(tabulate(
        [
            (k, v)
            for k, v in sorted(get_versions().items())
            if "ols_version_date_format" not in registry[k] and "ols_version_type" not in registry[k]
        ],
        headers=['Prefix', 'Version'],
    ))


@main.command()
@click.pass_context
def update(ctx: click.Context):
    """Update the Bioregistry."""
    ctx.invoke(download)
    ctx.invoke(align)
    ctx.invoke(lint)
    ctx.invoke(copy)
    ctx.invoke(compare)
    ctx.invoke(curation)
    ctx.invoke(warnings)


if __name__ == '__main__':
    main()



# -*- coding: utf-8 -*-

"""Resolvers for CURIE (e.g., pairs of prefix and identifier)."""

from typing import Callable, Mapping, Optional, Sequence, Tuple

from .constants import BIOREGISTRY_REMOTE_URL
from .resolve import (
    get, get_banana, get_bioportal_prefix, get_identifiers_org_prefix, get_n2t_prefix, get_obofoundry_prefix,
    get_ols_prefix, get_pattern_re, namespace_in_lui, normalize_prefix,
)

__all__ = [
    'validate',
    'get_providers',
    'get_providers_list',
    'get_identifiers_org_url',
    'get_identifiers_org_curie',
    'get_obofoundry_link',
    'get_ols_link',
    'get_bioportal_url',
    'get_n2t_url',
    'get_link',
    'get_registry_resolve_url',
]


def validate(prefix: str, identifier: str) -> Optional[bool]:
    """Validate the identifier against the prefix's pattern, if it exists."""
    pattern = get_pattern_re(prefix)
    if pattern is None:
        return None
    return bool(pattern.match(_get_modified_id(prefix, identifier)))


def _get_modified_id(prefix: str, identifier: str) -> str:
    if not namespace_in_lui(prefix):
        return identifier
    banana = get_banana(prefix)
    if banana:
        if identifier.startswith(f'{banana}:'):
            return identifier
        else:
            return f'{banana}:{identifier}'
    else:
        if identifier.startswith(prefix.upper()):
            return identifier
        else:
            return f'{prefix.upper()}:{identifier}'


def get_default_url(prefix: str, identifier: str) -> Optional[str]:
    """Get the default URL for the given CURIE."""
    entry = get(prefix)
    if entry is None:
        return None
    url = entry.get('url')
    if url is None:
        return None
    return url.replace('$1', identifier)


def get_providers(prefix: str, identifier: str) -> Mapping[str, str]:
    """Get all providers for the CURIE."""
    return dict(get_providers_list(prefix, identifier))


def get_providers_list(prefix: str, identifier: str) -> Sequence[Tuple[str, str]]:
    """Get all providers for the CURIE."""
    rv = []
    for provider, get_url in PROVIDER_FUNCTIONS.items():
        link = get_url(prefix, identifier)
        if link is not None:
            rv.append((provider, link))
    if not rv:
        return rv

    bioregistry_link = _get_bioregistry_link(prefix, identifier)
    if not bioregistry_link:
        return rv

    # if a default URL is available, it goes first. otherwise the bioregistry URL goes first.
    rv.insert(1 if rv[0][0] == 'default' else 0, ('bioregistry', bioregistry_link))
    return rv


def get_identifiers_org_url(prefix: str, identifier: str) -> Optional[str]:
    """Get the identifiers.org URL for the given CURIE."""
    curie = get_identifiers_org_curie(prefix, identifier)
    if curie is None:
        return None
    return f'https://identifiers.org/{curie}'


def get_n2t_url(prefix: str, identifier: str) -> Optional[str]:
    """Get the name-to-thing URL for the given CURIE."""
    n2t_prefix = get_n2t_prefix(prefix)
    if n2t_prefix is None:
        return None
    curie = f'{n2t_prefix}:{identifier}'
    if curie is None:
        return None
    return f'https://n2t.net/{curie}'


def get_bioportal_url(prefix: str, identifier: str) -> Optional[str]:
    """Get the Bioportal URL for the given CURIE.

    :param prefix: The prefix in the CURIE
    :param identifier: The identifier in the CURIE
    :return: A link to the Bioportal page

    >>> get_bioportal_url('chebi', '24431')
    "https://bioportal.bioontology.org/ontologies/CHEBI/?p=classes&conceptid=http%3A%2F%2Fpurl.obolibrary.org%2Fobo%2FCHEBI_24431"
    """
    bioportal_prefix = get_bioportal_prefix(prefix)
    if bioportal_prefix is None:
        return None
    obo_link = get_obofoundry_link(prefix, identifier)
    if obo_link is not None:
        return f'https://bioportal.bioontology.org/ontologies/{bioportal_prefix}/?p=classes&conceptid={obo_link}'
    # TODO there must be other rules?
    return None


# MIRIAM definitions that don't make any sense
MIRIAM_BLACKLIST = {
    # this one uses the names instead of IDs, and points to a dead resource.
    # See https://github.com/identifiers-org/identifiers-org.github.io/issues/139
    'pid.pathway',
}


def get_identifiers_org_curie(prefix: str, identifier: str) -> Optional[str]:
    """Get the identifiers.org CURIE for the given CURIE."""
    miriam_prefix = get_identifiers_org_prefix(prefix)
    if miriam_prefix is None or miriam_prefix in MIRIAM_BLACKLIST:
        return None
    banana = get_banana(prefix)
    if banana:
        if identifier.startswith(f'{banana}:'):
            return identifier
        else:
            return f'{banana}:{identifier}'
    elif namespace_in_lui(prefix):
        if identifier.startswith(prefix.upper()):
            return identifier
        else:
            return f'{prefix.upper()}:{identifier}'
    else:
        return f'{miriam_prefix}:{identifier}'


def get_obofoundry_link(prefix: str, identifier: str) -> Optional[str]:
    """Get the OBO Foundry URL if possible."""
    obo_prefix = get_obofoundry_prefix(prefix)
    if obo_prefix is None:
        return None
    return f'http://purl.obolibrary.org/obo/{obo_prefix.upper()}_{identifier}'


def get_ols_link(prefix: str, identifier: str) -> Optional[str]:
    """Get the OLS URL if possible."""
    ols_prefix = get_ols_prefix(prefix)
    obo_link = get_obofoundry_link(prefix, identifier)
    if ols_prefix is None or obo_link is None:
        return None
    return f'https://www.ebi.ac.uk/ols/ontologies/{ols_prefix}/terms?iri={obo_link}'


def _get_bioregistry_link(prefix: str, identifier: str) -> Optional[str]:
    """Get the bioregistry link.

    :param prefix: The prefix in the CURIE
    :param identifier: The identifier in the CURIE
    :return: A link to the bioregistry resolver

    >>> _get_bioregistry_link('chebi', '1234')
    http://bioregistry.io/chebi:1234
    """
    norm_prefix = normalize_prefix(prefix)
    if norm_prefix is None:
        return None
    return f'{BIOREGISTRY_REMOTE_URL.rstrip()}/{norm_prefix}:{identifier}'


PROVIDER_FUNCTIONS: Mapping[str, Callable[[str, str], Optional[str]]] = {
    'default': get_default_url,
    'miriam': get_identifiers_org_url,
    'obofoundry': get_obofoundry_link,
    'ols': get_ols_link,
    'n2t': get_n2t_url,
    'bioportal': get_bioportal_url,
}

LINK_PRIORITY = [
    'default',
    'bioregistry',
    'miriam',
    'ols',
    'obofoundry',
    'n2t',
    'bioportal',
]


def get_link(prefix: str, identifier: str, use_bioregistry_io: bool = True) -> Optional[str]:
    """Get the best link for the CURIE, if possible."""
    providers = get_providers(prefix, identifier)
    for key in LINK_PRIORITY:
        if not use_bioregistry_io and key == 'bioregistry':
            continue
        if key not in providers:
            continue
        rv = providers[key]
        if rv is not None:
            return rv
    return None


def get_registry_resolve_url(metaprefix: str, prefix: str, identifier: str) -> Optional[str]:
    """Get the URL to resolve the given prefix/identifier pair with the given resolver."""
    providers = get_providers(prefix, identifier)
    if not providers:
        return None
    return providers.get(metaprefix)



# -*- coding: utf-8 -*-

"""Align the Prefix Commons with the Bioregistry."""

from typing import Any, Dict, Sequence

from bioregistry.align.utils import Aligner
from bioregistry.external.prefix_commons import get_prefix_commons

__all__ = [
    'PrefixCommonsAligner',
]


class PrefixCommonsAligner(Aligner):
    """Aligner for Prefix Commons."""

    key = 'prefixcommons'
    getter = get_prefix_commons
    curation_header = ['formatter', 'identifiers', 'purl']

    def prepare_external(self, external_id, external_entry) -> Dict[str, Any]:
        """Prepare Prefix Commons data to be added to the Prefix Commons for each BioPortal registry entry."""
        formatter = external_entry['formatter'].strip()
        return {
            'formatter': formatter,
            'is_identifiers': formatter.startswith('http://identifiers.org'),
            'is_obo': formatter.startswith('http://purl.obolibrary.org'),
        }

    def get_curation_row(self, external_id, external_entry) -> Sequence[str]:
        """Prepare curation rows for unaligned Prefix Commons registry entries."""
        formatter = external_entry['formatter'].strip()
        return [
            formatter,
            formatter.startswith('http://identifiers.org'),
            formatter.startswith('http://purl.obolibrary.org'),
        ]


if __name__ == '__main__':
    PrefixCommonsAligner.align()



# -*- coding: utf-8 -*-

"""Make the curation list."""

import os

import click
import yaml

from bioregistry import get_example, get_format, get_name, get_pattern, read_bioregistry
from bioregistry.constants import DOCS_DATA

items = sorted(read_bioregistry().items())


def _g(predicate):
    return [
        {
            'prefix': bioregistry_id,
            'name': get_name(bioregistry_id),
        }
        for bioregistry_id, bioregistry_entry in items
        if predicate(bioregistry_id, bioregistry_entry)
    ]


@click.command()
def curation():
    """Make curation list."""
    missing_wikidata_database = _g(lambda prefix, entry: entry.get('wikidata', {}).get('database') is None)
    missing_pattern = _g(lambda prefix, entry: get_pattern(prefix) is None)
    missing_format_url = _g(lambda prefix, entry: get_format(prefix) is None)
    missing_example = _g(lambda prefix, entry: get_example(prefix) is None)

    with open(os.path.join(DOCS_DATA, 'curation.yml'), 'w') as file:
        yaml.safe_dump(
            {
                'wikidata': missing_wikidata_database,
                'pattern': missing_pattern,
                'formatter': missing_format_url,
                'example': missing_example,
            },
            file,
        )


if __name__ == '__main__':
    curation()



# -*- coding: utf-8 -*-

"""Download registry information from NCBI."""

import json
import re
from typing import Dict
from urllib.parse import urlsplit, urlunsplit

import click
from bs4 import BeautifulSoup

from bioregistry.constants import BIOREGISTRY_MODULE

URL = 'https://www.ncbi.nlm.nih.gov/genbank/collab/db_xref/'
NCBI_URL_PARTS = urlsplit(URL)
DATA_TABLE_CAPTION_RE = re.compile(r'db_xref List')

MISSING = {
    'EnsemblGenomes-Gn',
}
REWRITES = {
    'JGI Phytozome': 'Phytozome',
    'UniProtKB/Swiss-Prot': 'UniProt',
}
XREF_PREFIX = [
    '/db_xref="',  # standard
    'db_xref="',  # ERIC
    '/dbxref="',  # PDB
    '/db_xref=',  # Xenbase
]
REDUNDANT = {
    'ATCC(dna)': 'ATCC',
    'ATCC(in host)': 'ATCC',
    'UniProtKB/TrEMBL': 'UniProt',
    'LocusID': 'GeneID',  # As of March 2005
}
OBSOLETE = {
    'ERIC',  # should be at http://www.ericbrc.org/portal/eric/, but gone.
    'PBmice',  # website down
}


def get_ncbi(force: bool = False) -> Dict[str, Dict[str, str]]:
    """Get the NCBI data."""
    parsed_path = BIOREGISTRY_MODULE.join(name='ncbi.json')
    if parsed_path.exists() and not force:
        with parsed_path.open() as file:
            return json.load(file)

    source_path = BIOREGISTRY_MODULE.ensure(url=URL, name='ncbi.html', force=force)
    with open(source_path) as f:
        soup = BeautifulSoup(f, 'html.parser')
    # find the data table based on its caption element
    data_table = soup.find('caption', string=DATA_TABLE_CAPTION_RE).parent

    rv = {}
    for row in data_table.find_all('tr'):
        cells = row.find_all('td')
        if not cells:
            continue

        prefix = cells[0].text.strip()
        prefix = REWRITES.get(prefix, prefix)

        if prefix in REDUNDANT:
            print('skipping', prefix)
            continue

        name = cells[2].text.strip()
        if not (prefix and name):  # blank line
            continue

        item = {'name': name}

        link = cells[0].find('a')
        if link and 'href' in link.attrs:
            link_href = link.attrs['href'].strip()
            if link_href:
                url_parts = urlsplit(link_href)
                if not url_parts.netloc:  # handle relative links
                    if url_parts.path.startswith('/'):  # relative to site root
                        url_parts = (
                            NCBI_URL_PARTS.scheme,
                            NCBI_URL_PARTS.netloc,
                            url_parts.path,
                            url_parts.query,
                            url_parts.fragment,
                        )
                    else:  # relative to the page we got it from
                        url_parts = (
                            NCBI_URL_PARTS.scheme,
                            NCBI_URL_PARTS.netloc,
                            NCBI_URL_PARTS.path + url_parts.path,
                            url_parts.query,
                            url_parts.fragment,
                        )
                    item['homepage'] = urlunsplit(url_parts)
                else:  # absolute links can pass through unchanged
                    item['homepage'] = link_href

        if prefix not in MISSING:
            examples = cells[4].text.strip()
            # only bother with the first line of examples
            example = examples.split('\n')[0].strip()

            for xref_prefix in XREF_PREFIX:
                if example.startswith(xref_prefix):
                    stripped_example = example[len(xref_prefix):].rstrip('"').lstrip()
                    break
            else:
                raise ValueError(f'[{prefix}] wrong xref prefix: {example}')

            if ':' not in stripped_example:
                raise ValueError(f'[{prefix}] not formatted as CURIE: {example}')

            if not stripped_example.lower().startswith(prefix.lower()):
                raise ValueError(f'[{prefix}] invalid prefix: {example}')

            parsed_prefix, identifier = [x.strip() for x in stripped_example.split(':', 1)]
            if prefix.casefold() != REWRITES.get(parsed_prefix, parsed_prefix).casefold():
                raise ValueError(f'example does not start with prefix {prefix} -> {identifier} from {example}')

            item['example'] = identifier

        rv[prefix] = item

    with parsed_path.open('w') as file:
        json.dump(rv, file, indent=2)

    return rv


@click.command()
def main():
    """Reload NCBI data."""
    get_ncbi(force=True)


if __name__ == '__main__':
    main()



# -*- coding: utf-8 -*-

"""Download the Gene Ontology registry."""

import json
import os
from typing import Optional

import click
import yaml

from bioregistry.constants import BIOREGISTRY_MODULE
from bioregistry.external.utils import list_to_map

__all__ = [
    'GO_FULL_PATH',
    'GO_URL',
    'get_go',
]

# Xrefs from GO that aren't generally useful
SKIP = {
    'TO_GIT',
    'OBO_SF_PO',
    'OBO_SF2_PO',
    'OBO_SF2_PECO',
    'PECO_GIT',
    'PO_GIT',
    'PSO_GIT',
    'EO_GIT',
}

# The key is redundant of the value
REDUNDANT = {
    "AspGD": "AspGD_LOCUS",
}

GO_YAML_PATH = BIOREGISTRY_MODULE.join(name='go.yml')
GO_FULL_PATH = BIOREGISTRY_MODULE.join(name='go.json')
GO_URL = 'https://raw.githubusercontent.com/geneontology/go-site/master/metadata/db-xrefs.yaml'


def get_go(
    cache_path: Optional[str] = GO_FULL_PATH,
    mappify: bool = False,

    force_download: bool = False,
):
    """Get the GO registry."""
    if not force_download and cache_path is not None and os.path.exists(cache_path):
        with open(cache_path) as file:
            entries = json.load(file)
    else:
        with BIOREGISTRY_MODULE.ensure(url=GO_URL).open() as file:
            entries = yaml.full_load(file)
        entries = [
            entry
            for entry in entries
            if entry['database'] not in SKIP and entry['database'] not in REDUNDANT
        ]
        if cache_path is not None:
            with open(cache_path, 'w') as file:
                json.dump(entries, file, indent=2, sort_keys=True)

    if mappify:
        entries = list_to_map(entries, 'database')

    return entries


@click.command()
def main():
    """Reload the GO data."""
    get_go(force_download=True)


if __name__ == '__main__':
    main()



# -*- coding: utf-8 -*-

"""Align the OBO Foundry with the Bioregistry."""

import requests
import requests.exceptions
import requests_ftp

from ..external import get_obofoundry
from ..utils import norm, secho, updater

requests_ftp.monkeypatch_session()
session = requests.Session()

OBO_KEYS = {
    'id',
    'prefix',
    'pattern',
    'namespaceEmbeddedInLui',
    'name',
    'deprecated',
    'description',
    'homepage',
}


def _prepare_obo(obofoundry_entry):  # noqa:C901
    prefix = obofoundry_entry['id']
    rv = {
        'prefix': prefix,
        'name': obofoundry_entry['title'],
        'deprecated': obofoundry_entry.get('is_obsolete', False),
    }

    license_dict = obofoundry_entry.get('license')
    if license_dict is not None:
        rv['license'] = license_dict['label']

    homepage = obofoundry_entry.get('homepage')
    if homepage is not None:
        rv['homepage'] = homepage

    contact_dict = obofoundry_entry.get('contact')
    if contact_dict is not None and contact_dict.get('email'):
        rv.update({
            'contact': contact_dict['email'],
            'contact.label': contact_dict['label'],
        })

    build = obofoundry_entry.get('build')
    if build is not None:
        method = build.get('method')
        if method is None and 'checkout in build':
            method = 'vcs'
        if method is None:
            secho(f'[{prefix}] missing method: {build}', fg='red')
            return rv

        if method == 'vcs':
            if 'system' not in build:
                secho(f'[{prefix}] missing build system', fg='red')
                return rv
            if build['system'] != 'git':
                secho(f'[{prefix}] unrecognized build system: {build["system"]}', fg='red')
                return rv
            checkout = build['checkout'].replace('  ', ' ')
            if not checkout.startswith('git clone https://github.com/'):
                secho(f'[{prefix}] unhandled build checkout: {checkout}', fg='red')
                return rv

            owner, repo = checkout.removeprefix('git clone https://github.com/').removesuffix('.git').split('/')
            rv['repo'] = f'https://github.com/{owner}/{repo}.git'

            path = build.get('path', '.')
            if path == '.':
                obo_url = f'https://raw.githubusercontent.com/{owner}/{repo}/master/{prefix}.obo'
            else:  # disregard the path since most repos don't actually use it anyway
                # TODO maybe try recovering if this doesn't work
                obo_url = f'https://raw.githubusercontent.com/{owner}/{repo}/master/{prefix}.obo'

            res = session.get(obo_url)
            if res.status_code == 200:
                rv['download.obo'] = obo_url
            else:
                secho(f"[{prefix}] [http {res.status_code}] see {rv['repo']} [{path}]", bold=True, fg='red')

        elif method == 'owl2obo':
            source_url = build['source_url']

            # parse repo if possible
            for url_prefix in ('https://github.com/', 'http://github.com/', 'https://raw.githubusercontent.com/'):
                if source_url.startswith(url_prefix):
                    owner, repo, *_ = source_url.removeprefix(url_prefix).split('/')
                    rv['repo'] = f'https://github.com/{owner}/{repo}.git'
                    break

            if source_url.endswith('.obo'):
                rv['download.obo'] = source_url
            elif source_url.endswith('.owl'):
                obo_url = source_url.removesuffix('.owl') + '.obo'
                res = session.get(obo_url)
                if res.status_code == 200:
                    rv['download.obo'] = source_url
                else:
                    secho(f'[{prefix}] [http {res.status_code}] problem with {obo_url}', bold=True, fg='red')
            else:
                secho(f'[{prefix}] unhandled build.source_url: {source_url}', fg='red')

        elif method == 'obo2owl':
            source_url = build['source_url']
            if source_url.endswith('.obo'):
                res = session.get(source_url)
                if res.status_code == 200:
                    rv['download.obo'] = source_url
                else:
                    secho(f'[{prefix}] [http {res.status_code}] problem with {source_url}', bold=True, fg='red')
            else:
                secho(f'[{prefix}] unhandled extension {source_url}', bold=True, fg='red')
        else:
            secho(f'[{prefix}] unhandled build method: {method}', fg='red')

    return rv


@updater
def align_obofoundry(registry):
    """Update OBOFoundry references."""
    obofoundry_id_to_bioregistry_id = {
        entry['obofoundry']['prefix']: key
        for key, entry in registry.items()
        if 'obofoundry' in entry
    }
    obofoundry_registry = get_obofoundry(mappify=True)

    obofoundry_norm_prefix_to_prefix = {
        norm(obo_key): obo_key
        for obo_key in obofoundry_registry
    }
    for bioregistry_id, entry in registry.items():
        if 'obofoundry' in entry:
            continue
        obofoundry_id = obofoundry_norm_prefix_to_prefix.get(norm(bioregistry_id))
        if obofoundry_id is not None:
            entry['obofoundry'] = {'prefix': obofoundry_id}
            obofoundry_id_to_bioregistry_id[obofoundry_id] = bioregistry_id

    for obofoundry_prefix, obofoundry_entry in obofoundry_registry.items():
        # Get key by checking the miriam.id key
        bioregistry_id = obofoundry_id_to_bioregistry_id.get(obofoundry_prefix)
        if bioregistry_id is None:
            if obofoundry_entry.get('is_obsolete'):
                secho(f'[{obofoundry_prefix}] skipping deprecated. If needed, add manually later', fg='yellow')
                continue

            bioregistry_id = obofoundry_prefix
            registry[bioregistry_id] = {}
            secho(f'[{obofoundry_prefix}] added: {obofoundry_entry["title"]}', fg='green')

        try:
            registry[bioregistry_id]['obofoundry'] = _prepare_obo(obofoundry_entry)
        except requests.exceptions.ConnectionError as e:
            secho(f'[{bioregistry_id}] failed to get data: {e}', fg='red')
            continue

    return registry


if __name__ == '__main__':
    align_obofoundry()



# -*- coding: utf-8 -*-

"""Tests for the bioregistry client."""

import unittest
from typing import Iterable, Tuple

import bioregistry


class TestResolve(unittest.TestCase):
    """Tests for getting Bioregistry content."""

    def test_resolve(self):
        """Test prefixes can be resolved properly."""
        for expected, query in [
            ('ncbitaxon', 'ncbitaxon'),
            ('ncbitaxon', 'NCBITaxon'),
            ('ncbitaxon', 'taxonomy'),
            ('bel', 'SCOMP'),
            ('bel', 'SFAM'),
            ('eccode', 'ec-code'),
            ('eccode', 'EC_CODE'),
            ('chembl.compound', 'chembl.compound'),
            ('chembl.compound', 'chemblcompound'),
            ('chembl', 'chembl'),
        ]:
            with self.subTest(query=query):
                self.assertEqual(expected, bioregistry.normalize_prefix(query))

    def test_get(self):
        """Test getting content from the bioregistry."""
        ncbitaxon_entry = bioregistry.get('ncbitaxon')
        self.assertIn('synonyms', ncbitaxon_entry)
        self.assertIn('NCBI_Taxon_ID', ncbitaxon_entry['synonyms'])
        self.assertIn('miriam', ncbitaxon_entry)
        self.assertIn('obofoundry', ncbitaxon_entry)
        self.assertIn('ols', ncbitaxon_entry)
        self.assertIn('wikidata', ncbitaxon_entry)

    def test_validate_none(self):
        """Test validation of identifiers for a prefix that does not exist."""
        self.assertIsNone(bioregistry.validate('lol', 'lol:nope'))

    def test_validate_true(self):
        """Test that validation returns true."""
        self.assert_validate([
            ('eccode', '1'),
            ('eccode', '1.1'),
            ('eccode', '1.1.1'),
            ('eccode', '1.1.1.1'),
            ('eccode', '1.1.123.1'),
            ('eccode', '1.1.1.123'),
            # Namespace in LUI: Standard rule for upper-casing
            ('chebi', '24867'),
            ('chebi', 'CHEBI:1234'),
            # BANANA
            ('vario', '0376'),  # this showcases the banana problem where the namespace in LUI is weird
            ('VariO', '0376'),
            ('did', 'sov:WRfXPg8dantKVubE3HX8pw'),
            ('did', 'did:sov:WRfXPg8dantKVubE3HX8pw'),
            ('go.ref', '0000041'),
            ('go.ref', 'GO_REF:0000041'),
        ])

    def assert_validate(self, examples: Iterable[Tuple[str, str]]) -> None:
        """Validate the examples."""
        for prefix, identifier in examples:
            is_valid = bioregistry.validate(prefix, identifier)
            if is_valid is False:
                with self.subTest(prefix=prefix, identifier=identifier):
                    self.fail(msg=f'CURIE {prefix}:{identifier}')

    def test_validate_false(self):
        """Test that validation returns false."""
        for prefix, identifier in [
            ('chebi', 'A1234'),
            ('chebi', 'chebi:1234'),
        ]:
            with self.subTest(prefix=prefix, identifier=identifier):
                self.assertFalse(bioregistry.validate(prefix, identifier))

    def test_lui(self):
        """Test the LUI makes sense (spoilers, they don't).

        Discussion is ongoing at:

        - https://github.com/identifiers-org/identifiers-org.github.io/issues/151
        """
        for prefix in bioregistry.read_bioregistry():
            if not bioregistry.namespace_in_lui(prefix):
                continue
            if bioregistry.get_banana(prefix):
                continue  # rewrite rules are applied to prefixes with bananas
            if prefix in {'ark', 'obi'}:
                continue  # these patterns on identifiers.org are garb
            with self.subTest(prefix=prefix):
                re_pattern = bioregistry.get_pattern(prefix)
                miriam_prefix = bioregistry.get_identifiers_org_prefix(prefix)
                self.assertTrue(
                    re_pattern.startswith(f'^{miriam_prefix.upper()}') or re_pattern.startswith(miriam_prefix.upper()),
                    msg=f'{prefix} pattern: {re_pattern}',
                )



# -*- coding: utf-8 -*-

"""Align Wikidata with the Bioregistry."""

from bioregistry.utils import query_wikidata, updater


@updater
def align_wikidata(registry):
    """Update Wikidata references."""
    properties = {}
    for bioregistry_id, v in registry.items():
        wikidata_property = v.get('wikidata', {}).get('property')
        if wikidata_property is not None:
            properties[wikidata_property] = bioregistry_id

    query = f'''
    SELECT
      ?prop ?propLabel ?format ?pattern ?homepage ?database ?databaseLabel
    WHERE
    {{
        OPTIONAL {{ ?prop wdt:P1921 ?format }}
        OPTIONAL {{ ?prop wdt:P1793 ?pattern }}
        OPTIONAL {{ ?prop wdt:P1896 ?homepage }}
        OPTIONAL {{ ?prop wdt:P1629 ?database }}
        VALUES ?prop {{{' '.join(set(f'wd:{p}' for p in properties))}}}
        SERVICE wikibase:label {{ bd:serviceParam wikibase:language "[AUTO_LANGUAGE],en". }}
    }}
    '''

    for entry in query_wikidata(query):
        entry = {k: v['value'] for k, v in entry.items()}
        if 'database' in entry:
            entry['database'] = entry.pop('database').removeprefix('http://www.wikidata.org/entity/')
        if 'databaseLabel' in entry:
            entry['database.label'] = entry.pop('databaseLabel')
        entry['property.label'] = entry.pop('propLabel')
        bioregistry_id = properties[entry.pop('prop').removeprefix('http://www.wikidata.org/entity/')]
        registry[bioregistry_id]['wikidata'].update(entry)

    return registry


if __name__ == '__main__':
    align_wikidata()



# -*- coding: utf-8 -*-

"""Web application for the Bioregistry."""

from flasgger import Swagger
from flask import Blueprint, Flask, abort, jsonify, render_template, request
from flask_bootstrap import Bootstrap

import bioregistry
from bioregistry.app.ui import ui_blueprint
from .utils import _autocomplete, _get_identifier, _normalize_prefix_or_404, _search
from ..resolve_identifier import _get_bioregistry_link

app = Flask(__name__)
Swagger.DEFAULT_CONFIG.update({
    'title': 'Bioregistry API',
    'description': 'A service for resolving CURIEs',
    'contact': {
        'responsibleDeveloper': 'Charles Tapley Hoyt',
        'email': 'cthoyt@gmail.com',
    },
    'version': '1.0',
})
Swagger(app)
Bootstrap(app)

api_blueprint = Blueprint('api', __name__, url_prefix='/api')


@api_blueprint.route('/registry/')
def resources():
    """List the entire Bioregistry."""
    return jsonify(bioregistry.read_bioregistry())


@api_blueprint.route('/metaregistry/')
def metaresources():
    """List the entire Bioregistry metaregistry."""
    return jsonify(bioregistry.read_metaregistry())


@api_blueprint.route('/metaregistry/<metaprefix>')
def metaresource(metaprefix: str):
    """List the registry."""
    data = bioregistry.get_registry(metaprefix)
    if not data:
        abort(404, f'Invalid metaprefix: {metaprefix}')
    return jsonify(data)


@api_blueprint.route('/registry/<prefix>')
def resource(prefix: str):
    """Get an entry.

    ---
    parameters:
    - name: prefix
      in: path
      description: The prefix for the entry
      required: true
      type: string
      example: doid
    """  # noqa:DAR101,DAR201
    prefix = _normalize_prefix_or_404(prefix)
    return jsonify(prefix=prefix, **bioregistry.get(prefix))  # type:ignore


@api_blueprint.route('/reference/<prefix>:<identifier>')
def reference(prefix: str, identifier: str):
    """Look up information on the CURIE.

    ---
    parameters:
    - name: prefix
      in: path
      description: The prefix for the entry
      required: true
      type: string
      example: efo
    - name: identifier
      in: path
      description: The identifier for the entry
      required: true
      type: string
      example: 0000311
    """  # noqa:DAR101,DAR201
    return jsonify(_get_identifier(prefix, identifier))


@api_blueprint.route('/search')
def search():
    """Search for a prefix.

    ---
    parameters:
    - name: q
      in: query
      description: The prefix for the entry
      required: true
      type: string
    """  # noqa:DAR101,DAR201
    q = request.args.get('q')
    if q is None:
        abort(400)
    return jsonify(_search(q))


@api_blueprint.route('/autocomplete')
def autocomplete():
    """Complete a resolution query.

    ---
    parameters:
    - name: q
      in: query
      description: The prefix for the entry
      required: true
      type: string
    """  # noqa:DAR101,DAR201
    q = request.args.get('q')
    if q is None:
        abort(400)
    return jsonify(_autocomplete(q))


app.register_blueprint(api_blueprint)
app.register_blueprint(ui_blueprint)


@app.route('/')
def home():
    """Render the homepage."""
    example_prefix, example_identifier = 'chebi', '138488'
    example_url = _get_bioregistry_link(example_prefix, example_identifier)
    return render_template(
        'home.html',
        example_url=example_url,
        example_prefix=example_prefix,
        example_identifier=example_identifier,
        registry_size=len(bioregistry.read_bioregistry()),
        metaregistry_size=len(bioregistry.read_metaregistry()),
    )


@app.route('/summary')
def summary():
    """Render the summary page."""
    return render_template('summary.html')


@app.route('/download')
def download():
    """Render the download page."""
    return render_template('download.html')


if __name__ == '__main__':
    app.run(debug=True)  # noqa



# -*- coding: utf-8 -*-

"""Download registry information from N2T."""

import itertools as itt
from operator import itemgetter

import click
import pystow
import yaml
from more_itertools import pairwise


URL = 'https://n2t.net/e/cdl_ebi_prefixes.yaml'

PATH = pystow.join('bioregistry', name='n2t.json')


def _parse_1(file):
    lines = (
        line.strip()
        for line in file
        if not line.startswith('#') and line.strip()
    )
    it = itt.groupby(lines, lambda line: line.startswith('-'))
    for a, grouped_lines in it:
        if a:
            k, v = [part.strip() for part in list(grouped_lines)[0].lstrip('-').lstrip().split(':', 1)]
            yield [(k, v)]
        else:
            yield [
                [part.strip() for part in line.split(':', 1)]
                for line in grouped_lines
            ]


def _parse(file):
    for a, b in pairwise(_parse_1(file)):
        yield dict(itt.chain(a, b))


def get_n2t():
    """Get the N2T registry."""
    path = pystow.ensure('bioregistry', url=URL)
    # they give malformed YAML so time to write a new parser
    with open(path) as file:
        rv = sorted(_parse(file), key=itemgetter('namespace'))

    nrv = {
        prefix: _clean_providers(lines)
        for prefix, lines in itt.groupby(rv, key=itemgetter('namespace'))
    }

    with PATH.open('w') as file:
        yaml.dump(nrv, file)
    return nrv


def _clean_providers(lines):
    providers = list(lines)
    for provider in providers:
        del provider['namespace']
    return providers


@click.command()
def main():
    """Reload the N2T data."""
    get_n2t()


if __name__ == '__main__':
    main()



# -*- coding: utf-8 -*-

"""Extract registry information."""

from .resolve import (  # noqa
    get, get_banana, get_description, get_email, get_example, get_format, get_homepage, get_identifiers_org_prefix,
    get_mappings, get_name, get_obo_download, get_owl_download, get_pattern, get_pattern_re, get_registry,
    get_registry_description, get_registry_homepage, get_registry_name, get_registry_url, get_synonyms, get_version,
    get_versions, has_terms, is_deprecated, namespace_in_lui, normalize_prefix, parse_curie,
)
from .resolve_identifier import (# noqa
    get_bioportal_url, get_identifiers_org_curie, get_identifiers_org_url, get_link, get_n2t_url, get_obofoundry_link,
    get_ols_link, get_providers, get_providers_list, get_registry_resolve_url, validate,
)
from .utils import read_bioregistry, read_metaregistry  # noqa



# -*- coding: utf-8 -*-

"""Align the OLS with the Bioregistry."""

import logging
from email.utils import parseaddr

from bioregistry.external import get_ols
from bioregistry.utils import norm, secho, updater

logger = logging.getLogger(__name__)


def _prepare_ols(ols_entry):
    ols_id = ols_entry['ontologyId']
    config = ols_entry['config']

    license_value = config['annotations'].get('license', [None])[0]
    if license_value in {'Unspecified', 'Unspecified'}:
        license_value = None

    rv = {
        'prefix': ols_id,
        'name': config['title'],
        'download': config['fileLocation'],
        'version': config['version'],
        'version.iri': config['versionIri'],
        'description': config['description'],
        'homepage': config['homepage'],
        'license': license_value,
    }

    email = config.get('mailingList')
    if email:
        name, email = parseaddr(email)
        if email.startswith('//'):
            logger.warning('[%s] invalid email address: %s', ols_id, config['mailingList'])
        else:
            rv['contact'] = email

    rv = {k: v for k, v in rv.items() if v}
    return rv


@updater
def align_ols(registry):
    """Update OLS references."""
    ols_id_to_bioregistry_id = {
        entry['ols']['prefix']: key
        for key, entry in registry.items()
        if 'ols' in entry
    }

    ols_registry = get_ols(mappify=True)

    ols_norm_prefix_to_prefix = {
        norm(obo_key): obo_key
        for obo_key in ols_registry
    }
    for bioregistry_id, entry in registry.items():
        if 'ols' in entry:
            continue
        ols_id = ols_norm_prefix_to_prefix.get(norm(bioregistry_id))
        if ols_id is not None:
            entry['ols'] = {'prefix': ols_id}
            ols_id_to_bioregistry_id[ols_id] = bioregistry_id

    for ols_prefix, ols_entry in ols_registry.items():
        bioregistry_id = ols_id_to_bioregistry_id.get(ols_prefix)
        if bioregistry_id is None:
            bioregistry_id = ols_prefix
            registry[bioregistry_id] = {}
            secho(f'[{ols_prefix}] added: {ols_entry["config"]["title"]}', fg='green')

        registry[bioregistry_id]['ols'] = _prepare_ols(ols_entry)

    return registry


if __name__ == '__main__':
    align_ols()



# -*- coding: utf-8 -*-

"""Test the resolution API."""

import sys
from functools import partial

import click
import requests
from tqdm import tqdm

import bioregistry

SLASH_URL_ENCODED = '%2F'


@click.command()
@click.option('-u', '--url', default='https://bioregistry.io', show_default=True)
@click.option('-l', '--local', is_flag=True)
def main(url: str, local: bool):
    """Test the API."""
    url = url.rstrip('/')
    if local:
        url = 'http://localhost:5000'
    click.echo(f'Testing resolution API on {url}')
    failure = False
    prefixes = tqdm(bioregistry.read_bioregistry())

    for prefix in prefixes:
        identifier = bioregistry.get_example(prefix)
        if identifier is None:
            continue
        prefixes.set_postfix({'prefix': prefix})
        req_url = f'{url}/{prefix}:{identifier}'
        res = requests.get(req_url, allow_redirects=False)
        log = partial(_log, req_url=req_url)
        if res.status_code == 302:  # redirect
            continue
        elif res.status_code != 404:
            text = res.text.splitlines()[3][len('<p>'):-len('</p>')]
            log(f'HTTP {res.status_code}: {res.reason} {text}', fg='red')
        elif not bioregistry.get_providers(prefix, identifier):
            continue
        elif '/' in identifier or SLASH_URL_ENCODED in identifier:
            log('contains slash 🎩 🎸', fg='red')
        elif not bioregistry.validate(prefix, identifier):
            pattern = bioregistry.get_pattern(prefix)
            if bioregistry.get_banana(prefix):
                log(f'banana {pattern} 🍌', fg='red')
            else:
                log(f'invalid example does not match pattern {pattern}', fg='red')
        else:
            log('404 unknown issue', fg='red')

        failure = True

    return sys.exit(1 if failure else 0)


def _log(s: str, req_url: str, **kwargs) -> None:
    with tqdm.external_write_mode(file=sys.stdout):
        click.secho(f'[{req_url}] {s}', **kwargs)


if __name__ == '__main__':
    main()



# -*- coding: utf-8 -*-

"""Align NCBI with the Bioregistry."""

import textwrap
from typing import List

from bioregistry.align.utils import Aligner
from bioregistry.external.ncbi import get_ncbi

__all__ = ['NcbiAligner']


class NcbiAligner(Aligner):
    """Aligner for NCBI xref registry."""

    key = 'ncbi'
    getter = get_ncbi
    curation_header = ('name', 'homepage', 'example')

    def get_curation_row(self, external_id, external_entry) -> List[str]:
        """Return the relevant fields from an NCBI entry for pretty-printing."""
        return [
            textwrap.shorten(external_entry['name'], 50),
            external_entry.get('example'),
            external_entry.get('homepage'),
        ]


if __name__ == '__main__':
    NcbiAligner.align()
